#!python
#cython: embedsignature=True, language_level=3

__all__ = ('BitmapFont',
           'BitmapFontException',
           '_SurfaceContainer',
           )

cimport cpython.pycapsule as pycapsule

import re
from os import path
from xml.etree.ElementTree import ElementTree, ParseError
from libc.string cimport memset
from kivy.core.image import ImageData


# ---------------------------------------------------------------------------
#    Definitions from SDL2 library
# ---------------------------------------------------------------------------

cdef extern from "SDL.h":
    ctypedef unsigned char Uint8
    ctypedef unsigned long Uint32
    ctypedef signed long Sint32
    ctypedef unsigned long long Uint64
    ctypedef signed long long Sint64
    ctypedef signed short Sint16
    ctypedef unsigned short Uint16
    ctypedef void *SDL_GLContext
    ctypedef Uint32 SDL_Keycode
    ctypedef Sint32 SDL_JoystickID

    int SDL_WINDOWPOS_UNDEFINED

    ctypedef enum:
        SDL_PIXELFORMAT_ARGB8888
        SDL_PIXELFORMAT_RGBA8888
        SDL_PIXELFORMAT_RGB888
        SDL_PIXELFORMAT_ABGR8888
        SDL_PIXELFORMAT_BGR888

    ctypedef enum SDL_GLattr:
        SDL_GL_RED_SIZE
        SDL_GL_GREEN_SIZE
        SDL_GL_BLUE_SIZE
        SDL_GL_ALPHA_SIZE
        SDL_GL_BUFFER_SIZE
        SDL_GL_DOUBLEBUFFER
        SDL_GL_DEPTH_SIZE
        SDL_GL_STENCIL_SIZE
        SDL_GL_ACCUM_RED_SIZE
        SDL_GL_ACCUM_GREEN_SIZE
        SDL_GL_ACCUM_BLUE_SIZE
        SDL_GL_ACCUM_ALPHA_SIZE
        SDL_GL_STEREO
        SDL_GL_MULTISAMPLEBUFFERS
        SDL_GL_MULTISAMPLESAMPLES
        SDL_GL_ACCELERATED_VISUAL
        SDL_GL_RETAINED_BACKING
        SDL_GL_CONTEXT_MAJOR_VERSION
        SDL_GL_CONTEXT_MINOR_VERSION
        SDL_GL_CONTEXT_EGL
        SDL_GL_CONTEXT_FLAGS
        SDL_GL_CONTEXT_PROFILE_MASK

    ctypedef enum SDL_BlendMode:
        SDL_BLENDMODE_NONE = 0x00000000
        SDL_BLENDMODE_BLEND = 0x00000001
        SDL_BLENDMODE_ADD = 0x00000002
        SDL_BLENDMODE_MOD = 0x00000004


    ctypedef enum SDL_TextureAccess:
        SDL_TEXTUREACCESS_STATIC
        SDL_TEXTUREACCESS_STREAMING
        SDL_TEXTUREACCESS_TARGET

    ctypedef enum SDL_RendererFlags:
        SDL_RENDERER_SOFTWARE = 0x00000001
        SDL_RENDERER_ACCELERATED = 0x00000002
        SDL_RENDERER_PRESENTVSYNC = 0x00000004

    ctypedef enum SDL_bool:
        SDL_FALSE = 0
        SDL_TRUE = 1

    cdef struct SDL_Rect:
        int x, y
        int w, h

    ctypedef struct SDL_Point:
        int x, y

    cdef struct SDL_Color:
        Uint8 r
        Uint8 g
        Uint8 b
        Uint8 unused

    cdef struct SDL_Palette:
        int ncolors
        SDL_Color *colors
        Uint32 version
        int refcount

    cdef struct SDL_PixelFormat:
        Uint32 format
        SDL_Palette *palette
        Uint8 BitsPerPixel
        Uint8 BytesPerPixel
        Uint8 padding[2]
        Uint32 Rmask
        Uint32 Gmask
        Uint32 Bmask
        Uint32 Amask
        Uint8 Rloss
        Uint8 Gloss
        Uint8 Bloss
        Uint8 Aloss
        Uint8 Rshift
        Uint8 Gshift
        Uint8 Bshift
        Uint8 Ashift
        int refcount
        SDL_PixelFormat *next


    cdef struct SDL_BlitMap

    cdef struct SDL_Surface:
        Uint32 flags
        SDL_PixelFormat *format
        int w, h
        int pitch
        void *pixels
        void *userdata
        int locked
        void *lock_data
        SDL_Rect clip_rect
        SDL_BlitMap *map
        int refcount


    ctypedef enum SDL_EventType:
        SDL_FIRSTEVENT     = 0,
        SDL_DROPFILE       = 0x1000,
        SDL_QUIT           = 0x100
        SDL_WINDOWEVENT    = 0x200
        SDL_SYSWMEVENT
        SDL_KEYDOWN        = 0x300
        SDL_KEYUP
        SDL_TEXTEDITING
        SDL_TEXTINPUT
        SDL_MOUSEMOTION     = 0x400
        SDL_MOUSEBUTTONDOWN = 0x401
        SDL_MOUSEBUTTONUP   = 0x402
        SDL_MOUSEWHEEL      = 0x403
        SDL_INPUTMOTION     = 0x500
        SDL_INPUTBUTTONDOWN
        SDL_INPUTBUTTONUP
        SDL_INPUTWHEEL
        SDL_INPUTPROXIMITYIN
        SDL_INPUTPROXIMITYOUT
        SDL_JOYAXISMOTION  = 0x600
        SDL_JOYBALLMOTION
        SDL_JOYHATMOTION
        SDL_JOYBUTTONDOWN
        SDL_JOYBUTTONUP
        SDL_FINGERDOWN      = 0x700
        SDL_FINGERUP
        SDL_FINGERMOTION
        SDL_TOUCHBUTTONDOWN
        SDL_TOUCHBUTTONUP
        SDL_DOLLARGESTURE   = 0x800
        SDL_DOLLARRECORD
        SDL_MULTIGESTURE
        SDL_CLIPBOARDUPDATE = 0x900
        SDL_EVENT_COMPAT1 = 0x7000
        SDL_EVENT_COMPAT2
        SDL_EVENT_COMPAT3
        SDL_USEREVENT    = 0x8000
        SDL_LASTEVENT    = 0xFFFF
        SDL_APP_TERMINATING
        SDL_APP_LOWMEMORY
        SDL_APP_WILLENTERBACKGROUND
        SDL_APP_DIDENTERBACKGROUND
        SDL_APP_WILLENTERFOREGROUND
        SDL_APP_DIDENTERFOREGROUND

    ctypedef enum SDL_WindowEventID:
        SDL_WINDOWEVENT_NONE           #< Never used */
        SDL_WINDOWEVENT_SHOWN          #< Window has been shown */
        SDL_WINDOWEVENT_HIDDEN         #< Window has been hidden */
        SDL_WINDOWEVENT_EXPOSED        #< Window has been exposed and should be
                                        #     redrawn */
        SDL_WINDOWEVENT_MOVED          #< Window has been moved to data1, data2
                                        # */
        SDL_WINDOWEVENT_RESIZED        #< Window has been resized to data1xdata2 */
        SDL_WINDOWEVENT_SIZE_CHANGED   #< The window size has changed, either as a result of an API call or through the system or user changing the window size. */
        SDL_WINDOWEVENT_MINIMIZED      #< Window has been minimized */
        SDL_WINDOWEVENT_MAXIMIZED      #< Window has been maximized */
        SDL_WINDOWEVENT_RESTORED       #< Window has been restored to normal size
                                        # and position */
        SDL_WINDOWEVENT_ENTER          #< Window has gained mouse focus */
        SDL_WINDOWEVENT_LEAVE          #< Window has lost mouse focus */
        SDL_WINDOWEVENT_FOCUS_GAINED   #< Window has gained keyboard focus */
        SDL_WINDOWEVENT_FOCUS_LOST     #< Window has lost keyboard focus */
        SDL_WINDOWEVENT_CLOSE           #< The window manager requests that the
                                        # window be closed */

    ctypedef enum SDL_RendererFlip:
        SDL_FLIP_NONE = 0x00000000
        SDL_FLIP_HORIZONTAL = 0x00000001
        SDL_FLIP_VERTICAL = 0x00000002

    ctypedef enum SDL_WindowFlags:
        SDL_WINDOW_FULLSCREEN = 0x00000001      #,         /**< fullscreen window */
        SDL_WINDOW_OPENGL = 0x00000002          #,             /**< window usable with OpenGL context */
        SDL_WINDOW_SHOWN = 0x00000004           #,              /**< window is visible */
        SDL_WINDOW_HIDDEN = 0x00000008          #,             /**< window is not visible */
        SDL_WINDOW_BORDERLESS = 0x00000010      #,         /**< no window decoration */
        SDL_WINDOW_RESIZABLE = 0x00000020       #,          /**< window can be resized */
        SDL_WINDOW_MINIMIZED = 0x00000040       #,          /**< window is minimized */
        SDL_WINDOW_MAXIMIZED = 0x00000080       #,          /**< window is maximized */
        SDL_WINDOW_INPUT_GRABBED = 0x00000100   #,      /**< window has grabbed input focus */
        SDL_WINDOW_INPUT_FOCUS = 0x00000200     #,        /**< window has input focus */
        SDL_WINDOW_MOUSE_FOCUS = 0x00000400     #,        /**< window has mouse focus */
        SDL_WINDOW_FOREIGN = 0x00000800         #            /**< window not created by SDL */
        SDL_WINDOW_FULLSCREEN_DESKTOP
        SDL_WINDOW_ALLOW_HIGHDPI

    cdef struct SDL_DropEvent:
        Uint32 type
        Uint32 timestamp
        char* file

    cdef struct SDL_MouseMotionEvent:
        Uint32 type
        Uint32 timestamp
        Uint32 windowID
        Uint32 which
        Uint32 state
        Sint32 x
        Sint32 y
        Sint32 xrel
        Sint32 yrel

    cdef struct SDL_MouseButtonEvent:
        Uint32 type
        Uint32 timestamp
        Uint32 windowID
        Uint32 which
        Uint8 button
        Uint8 state
        Uint8 clicks
        Sint32 x
        Sint32 y

    cdef struct SDL_WindowEvent:
        Uint32 type
        Uint32 timestamp
        Uint32 windowID
        Uint8 event
        Sint32 data1
        Sint32 data2

    ctypedef Sint64 SDL_TouchID
    ctypedef Sint64 SDL_FingerID

    cdef struct SDL_TouchFingerEvent:
        Uint32 type
        Uint32 windowID
        SDL_TouchID touchId
        SDL_FingerID fingerId
        float x
        float y
        float dx
        float dy
        float pressure

    cdef struct SDL_Keysym:
        SDL_Scancode scancode       # SDL physical key code - see ::SDL_Scancode for details */
        SDL_Keycode sym             # SDL virtual key code - see ::SDL_Keycode for details */
        Uint16 mod                  # current key modifiers */
        Uint32 unused

    cdef struct SDL_KeyboardEvent:
        Uint32 type         # ::SDL_KEYDOWN or ::SDL_KEYUP
        Uint32 timestamp
        Uint32 windowID     # The window with keyboard focus, if any
        Uint8 state         # ::SDL_PRESSED or ::SDL_RELEASED
        Uint8 repeat        # Non-zero if this is a key repeat
        SDL_Keysym keysym   # The key that was pressed or released

    cdef struct SDL_TextEditingEvent:
        Uint32 type                                 # ::SDL_TEXTEDITING */
        Uint32 timestamp
        Uint32 windowID                             # The window with keyboard focus, if any */
        char *text                                  # The editing text */
        Sint32 start                                # The start cursor of selected editing text */
        Sint32 length                               # The length of selected editing text */

    cdef struct SDL_TextInputEvent:
        Uint32 type                               # ::SDL_TEXTINPUT */
        Uint32 timestamp
        Uint32 windowID                           # The window with keyboard focus, if any */
        char *text                                # The input text */

    cdef struct SDL_MouseWheelEvent:
        Uint32 type
        Uint32 windowID
        int x
        int y
    cdef struct SDL_JoyAxisEvent:
        Uint32 type
        Uint32 timestamp
        SDL_JoystickID which
        Uint8 axis
        Sint16 value
    cdef struct SDL_JoyBallEvent:
        Uint32 type
        Uint32 timestamp
        SDL_JoystickID which
        Uint8 ball
        Sint16  xrel
        Sint16  yrel
    cdef struct SDL_JoyHatEvent:
        Uint32 type
        Uint32 timestamp
        SDL_JoystickID which
        Uint8 hat
        Uint8 value
    cdef struct SDL_JoyButtonEvent:
        Uint32 type
        Uint32 timestamp
        SDL_JoystickID which
        Uint8 button
        Uint8 state
    cdef struct SDL_QuitEvent:
        pass
    cdef struct SDL_UserEvent:
        Uint32 type
        Uint32 timestamp
        Uint32 windowID
        int code
        void *data1
        void *data2

    cdef struct SDL_SysWMEvent:
        pass
    cdef struct SDL_TouchButtonEvent:
        pass
    cdef struct SDL_MultiGestureEvent:
        pass
    cdef struct SDL_DollarGestureEvent:
        pass

    cdef union SDL_Event:
        Uint32 type
        SDL_WindowEvent window
        SDL_KeyboardEvent key
        SDL_TextEditingEvent edit
        SDL_TextInputEvent text
        SDL_MouseMotionEvent motion
        SDL_MouseButtonEvent button
        SDL_DropEvent drop
        SDL_MouseWheelEvent wheel
        SDL_JoyAxisEvent jaxis
        SDL_JoyBallEvent jball
        SDL_JoyHatEvent jhat
        SDL_JoyButtonEvent jbutton
        SDL_QuitEvent quit
        SDL_UserEvent user
        SDL_SysWMEvent syswm
        SDL_TouchFingerEvent tfinger
        SDL_TouchButtonEvent tbutton
        SDL_MultiGestureEvent mgesture
        SDL_DollarGestureEvent dgesture

    cdef struct SDL_RendererInfo:
        char *name
        Uint32 flags
        Uint32 num_texture_formats
        Uint32 texture_formats[16]
        int max_texture_width
        int max_texture_height

    ctypedef struct SDL_Texture
    ctypedef struct SDL_Renderer
    ctypedef struct SDL_Window
    ctypedef struct SDL_DisplayMode:
        Uint32 format
        int w
        int h
        int refresh_rate
        void *driverdata

    cdef struct SDL_RWops:
        long (* seek) (SDL_RWops * context, long offset,int whence)
        size_t(* read) ( SDL_RWops * context, void *ptr, size_t size, size_t maxnum)
        size_t(* write) (SDL_RWops * context, void *ptr,size_t size, size_t num)
        int (* close) (SDL_RWops * context)

    cdef enum SDL_Keymod:
        KMOD_NONE
        KMOD_LSHIFT
        KMOD_RSHIFT
        KMOD_LCTRL
        KMOD_RCTRL
        KMOD_LALT
        KMOD_RALT
        KMOD_LGUI
        KMOD_RGUI
        KMOD_NUM
        KMOD_CAPS
        KMOD_MODE
        KMOD_RESERVED

    ctypedef enum SDL_Scancode:
        pass

    ctypedef int SDL_EventFilter(void* userdata, SDL_Event* event)

    cdef char *SDL_HINT_ORIENTATIONS
    cdef char *SDL_HINT_VIDEO_WIN_D3DCOMPILER
    cdef char *SDL_HINT_ACCELEROMETER_AS_JOYSTICK

    cdef int SDL_QUERY               = -1
    cdef int SDL_IGNORE              =  0
    cdef int SDL_DISABLE             =  0
    cdef int SDL_ENABLE              =  1
    cdef int SDL_INIT_TIMER          = 0x00000001
    cdef int SDL_INIT_AUDIO          = 0x00000010
    cdef int SDL_INIT_VIDEO          = 0x00000020  # SDL_INIT_VIDEO implies SDL_INIT_EVENTS */
    cdef int SDL_INIT_JOYSTICK       = 0x00000200  # SDL_INIT_JOYSTICK implies SDL_INIT_EVENTS */
    cdef int SDL_INIT_HAPTIC         = 0x00001000
    cdef int SDL_INIT_GAMECONTROLLER = 0x00002000  # SDL_INIT_GAMECONTROLLER implies SDL_INIT_JOYSTICK */
    cdef int SDL_INIT_EVENTS         = 0x00004000
    cdef int SDL_INIT_NOPARACHUTE    = 0x00100000  # Don't catch fatal signals */

    cdef SDL_Renderer * SDL_CreateRenderer(SDL_Window * window, int index, Uint32 flags)
    cdef void SDL_DestroyRenderer (SDL_Renderer * renderer)
    cdef SDL_Texture * SDL_CreateTexture(SDL_Renderer * renderer, Uint32 format, int access, int w, int h)
    cdef SDL_Texture * SDL_CreateTextureFromSurface(SDL_Renderer * renderer, SDL_Surface * surface)
    cdef SDL_Surface * SDL_CreateRGBSurface(Uint32 flags, int width, int height, int depth, Uint32 Rmask, Uint32 Gmask, Uint32 Bmask, Uint32 Amask) nogil
    cdef int SDL_RenderCopy(SDL_Renderer * renderer, SDL_Texture * texture, SDL_Rect * srcrect, SDL_Rect * dstrect)
    cdef int SDL_RenderCopyEx(SDL_Renderer * renderer, SDL_Texture * texture, SDL_Rect * srcrect, SDL_Rect * dstrect, double angle, SDL_Point *center, SDL_RendererFlip flip)
    cdef void SDL_RenderPresent(SDL_Renderer * renderer)
    cdef SDL_bool SDL_RenderTargetSupported(SDL_Renderer *renderer)
    cdef int SDL_SetRenderTarget(SDL_Renderer *renderer, SDL_Texture *texture)
    cdef void SDL_DestroyTexture(SDL_Texture * texture)
    cdef void SDL_FreeSurface(SDL_Surface * surface) nogil
    cdef int SDL_SetSurfaceBlendMode(SDL_Surface * surface, int blendMode)
    cdef int SDL_SetSurfaceAlphaMod(SDL_Surface * surface, char alpha)
    cdef int SDL_UpperBlit (SDL_Surface * src, SDL_Rect * srcrect, SDL_Surface * dst, SDL_Rect * dstrect)
    cdef int SDL_BlitSurface(SDL_Surface * src, SDL_Rect * srcrect, SDL_Surface * dst, SDL_Rect * dstrect)
    cdef int SDL_LockTexture(SDL_Texture * texture, SDL_Rect * rect, void **pixels, int *pitch)
    cdef void SDL_UnlockTexture(SDL_Texture * texture)
    cdef void SDL_GetWindowSize(SDL_Window * window, int *w, int *h)
    cdef Uint32 SDL_GetWindowFlags(SDL_Window * window)
    cdef SDL_Window * SDL_CreateWindow(char *title, int x, int y, int w, int h, Uint32 flags)
    cdef void SDL_DestroyWindow (SDL_Window * window)
    cdef int SDL_SetRenderDrawColor(SDL_Renderer * renderer, Uint8 r, Uint8 g, Uint8 b, Uint8 a)
    cdef int SDL_RenderClear(SDL_Renderer * renderer)
    cdef int SDL_SetTextureBlendMode(SDL_Texture * texture, SDL_BlendMode blendMode)
    cdef int SDL_GetTextureBlendMode(SDL_Texture * texture, SDL_BlendMode *blendMode)
    cdef SDL_Surface * SDL_CreateRGBSurfaceFrom(void *pixels, int width, int height, int depth, int pitch, Uint32 Rmask, Uint32 Gmask, Uint32 Bmask, Uint32 Amask)
    cdef SDL_Surface* SDL_ConvertSurface(SDL_Surface* src, SDL_PixelFormat* fmt, Uint32 flags)
    cdef SDL_Surface* SDL_ConvertSurfaceFormat(SDL_Surface* src, Uint32
            pixel_format, Uint32 flags)
    cdef const char* SDL_GetPixelFormatName(Uint32 format)
    cdef int SDL_Init(Uint32 flags)
    cdef void SDL_Quit()
    cdef int SDL_EnableUNICODE(int enable)
    cdef Uint32 SDL_GetTicks()
    cdef void SDL_Delay(Uint32 ms) nogil
    cdef Uint8 SDL_EventState(Uint32 type, int state)
    cdef int SDL_PollEvent(SDL_Event * event) nogil
    cdef void SDL_SetEventFilter(SDL_EventFilter *filter, void* userdata)
    cdef SDL_RWops * SDL_RWFromFile(char *file, char *mode)
    cdef SDL_RWops * SDL_RWFromMem(void *mem, int size)
    cdef SDL_RWops * SDL_RWFromConstMem(void *mem, int size)
    cdef void SDL_FreeRW(SDL_RWops *area)
    cdef int SDL_GetRendererInfo(SDL_Renderer *renderer, SDL_RendererInfo *info)
    cdef int SDL_RenderSetViewport(SDL_Renderer * renderer, SDL_Rect * rect)
    cdef int SDL_GetCurrentDisplayMode(int displayIndex, SDL_DisplayMode * mode)
    cdef int SDL_GetDesktopDisplayMode(int displayIndex, SDL_DisplayMode * mode)
    cdef int SDL_SetTextureColorMod(SDL_Texture * texture, Uint8 r, Uint8 g, Uint8 b)
    cdef int SDL_SetTextureAlphaMod(SDL_Texture * texture, Uint8 alpha)
    cdef char * SDL_GetError()
    cdef SDL_bool SDL_SetHint(char *name, char *value)
    cdef Uint8 SDL_GetMouseState(int* x,int* y)
    cdef SDL_GLContext SDL_GL_CreateContext(SDL_Window* window)
    cdef int SDL_GetNumVideoDisplays()
    cdef int SDL_GetNumDisplayModes(int displayIndex)
    cdef int SDL_GetDisplayMode(int displayIndex, int index, SDL_DisplayMode * mode)
    cdef SDL_bool SDL_HasIntersection(SDL_Rect * A, SDL_Rect * B) nogil
    cdef SDL_bool SDL_IntersectRect(SDL_Rect * A, SDL_Rect * B, SDL_Rect * result) nogil
    cdef void SDL_UnionRect(SDL_Rect * A, SDL_Rect * B, SDL_Rect * result) nogil
    cdef Uint64 SDL_GetPerformanceCounter() nogil
    cdef Uint64 SDL_GetPerformanceFrequency() nogil
    cdef int SDL_GL_SetAttribute(SDL_GLattr attr, int value)
    cdef int SDL_GetNumRenderDrivers()
    cdef int SDL_GetRenderDriverInfo(int index, SDL_RendererInfo* info)
    cdef int SDL_GL_BindTexture(SDL_Texture *texture, float *texw, float *texh)
    cdef int SDL_GL_UnbindTexture(SDL_Texture *texture)
    cdef int SDL_RenderReadPixels(SDL_Renderer * renderer, SDL_Rect * rect, Uint32 format, void *pixels, int pitch) nogil
    cdef int SDL_PushEvent(SDL_Event * event) nogil
    cdef int SDL_WaitEvent(SDL_Event * event) nogil

    cdef void SDL_SetClipboardText(char * text)
    cdef const char * SDL_GetClipboardText()
    cdef SDL_bool SDL_HasClipboardText()
    cdef int SDL_GetNumVideoDrivers()
    cdef const char *SDL_GetVideoDriver(int index)
    cdef int SDL_VideoInit(const char *driver_name)
    cdef void SDL_VideoQuit()
    cdef const char *SDL_GetCurrentVideoDriver()
    cdef int SDL_GetNumVideoDisplays()
    cdef const char * SDL_GetDisplayName(int displayIndex)
    cdef int SDL_GetDisplayBounds(int displayIndex, SDL_Rect * rect)
    cdef int SDL_GetNumDisplayModes(int displayIndex)
    cdef int SDL_GetDesktopDisplayMode(int displayIndex, SDL_DisplayMode * mode)
    cdef int SDL_GetCurrentDisplayMode(int displayIndex, SDL_DisplayMode * mode)
    cdef SDL_DisplayMode * SDL_GetClosestDisplayMode(int displayIndex, const SDL_DisplayMode * mode, SDL_DisplayMode * closest)
    cdef int SDL_SetWindowDisplayMode(SDL_Window * window, SDL_DisplayMode * mode)
    cdef int SDL_GetWindowDisplayMode(SDL_Window * window, SDL_DisplayMode * mode)
    cdef int SDL_GetWindowDisplayIndex(SDL_Window * window)
    cdef Uint32 SDL_GetWindowPixelFormat(SDL_Window * window)
    cdef SDL_Window * SDL_CreateWindowFrom(const void *data)
    cdef Uint32 SDL_GetWindowID(SDL_Window * window)
    cdef SDL_Window * SDL_GetWindowFromID(Uint32 id)
    cdef Uint32 SDL_GetWindowFlags(SDL_Window * window)
    cdef void SDL_SetWindowTitle(SDL_Window * window, char *title)
    cdef const char *SDL_GetWindowTitle(SDL_Window * window)
    cdef void SDL_SetWindowIcon(SDL_Window * window, SDL_Surface *icon)
    cdef void* SDL_SetWindowData(SDL_Window * window, char *name, void *data)
    cdef void *SDL_GetWindowData(SDL_Window * window, char *name)
    cdef void SDL_SetWindowPosition(SDL_Window * window, int x, int y)
    cdef void SDL_GetWindowPosition(SDL_Window * window, int *x, int *y)
    cdef void SDL_SetWindowSize(SDL_Window * window, int w, int h)
    cdef void SDL_GetWindowSize(SDL_Window * window, int *w, int *h)
    cdef void SDL_SetWindowMinimumSize(SDL_Window * window, int min_w, int min_h)
    cdef void SDL_SetWindowBordered(SDL_Window * window, SDL_bool bordered)
    cdef void SDL_ShowWindow(SDL_Window * window)
    cdef int SDL_ShowCursor(int toggle)
    cdef void SDL_HideWindow(SDL_Window * window)
    cdef void SDL_RaiseWindow(SDL_Window * window)
    cdef void SDL_MaximizeWindow(SDL_Window * window)
    cdef void SDL_MinimizeWindow(SDL_Window * window)
    cdef void SDL_RestoreWindow(SDL_Window * window)
    cdef int SDL_SetWindowFullscreen(SDL_Window * window, SDL_bool fullscreen)
    cdef SDL_Surface * SDL_GetWindowSurface(SDL_Window * window)
    cdef int SDL_UpdateWindowSurface(SDL_Window * window)
    cdef void SDL_SetWindowGrab(SDL_Window * window, SDL_bool grabbed)
    cdef SDL_bool SDL_GetWindowGrab(SDL_Window * window)
    cdef int SDL_SetWindowBrightness(SDL_Window * window, float brightness)
    cdef float SDL_GetWindowBrightness(SDL_Window * window)
    cdef void SDL_DestroyWindow(SDL_Window * window)
    cdef SDL_bool SDL_IsScreenSaverEnabled()
    cdef void SDL_EnableScreenSaver()
    cdef void SDL_DisableScreenSaver()
    cdef int SDL_GL_LoadLibrary(const char *path)
    cdef void *SDL_GL_GetProcAddress(const char *proc)
    cdef void SDL_GL_UnloadLibrary()
    cdef int SDL_GL_SetAttribute(SDL_GLattr attr, int value)
    cdef int SDL_GL_GetAttribute(SDL_GLattr attr, int *value)
    cdef int SDL_GL_MakeCurrent(SDL_Window * window, SDL_GLContext context)
    cdef SDL_Window* SDL_GL_GetCurrentWindow()
    cdef SDL_GLContext SDL_GL_GetCurrentContext()
    cdef int SDL_GL_SetSwapInterval(int interval)
    cdef int SDL_GL_GetSwapInterval()
    cdef void SDL_GL_SwapWindow(SDL_Window * window)
    cdef void SDL_GL_DeleteContext(SDL_GLContext context)

    cdef SDL_Window * SDL_GetKeyboardFocus()
    cdef Uint8 *SDL_GetKeyboardState(int *numkeys)
    cdef SDL_Keymod SDL_GetModState()
    cdef void SDL_SetModState(SDL_Keymod modstate)
    cdef SDL_Keycode SDL_GetKeyFromScancode(SDL_Scancode scancode)
    cdef SDL_Scancode SDL_GetScancodeFromKey(SDL_Keycode key)
    cdef char *SDL_GetScancodeName(SDL_Scancode scancode)
    cdef SDL_Scancode SDL_GetScancodeFromName(char *name)
    cdef char *SDL_GetKeyName(SDL_Keycode key)
    cdef SDL_Keycode SDL_GetKeyFromName(char *name)
    cdef void SDL_StartTextInput()
    cdef SDL_bool SDL_IsTextInputActive()
    cdef void SDL_StopTextInput()
    cdef void SDL_SetTextInputRect(SDL_Rect *rect)
    cdef SDL_bool SDL_HasScreenKeyboardSupport()
    cdef SDL_bool SDL_IsScreenKeyboardShown(SDL_Window *window)
    cdef void SDL_GL_GetDrawableSize(SDL_Window *window, int *w, int *h)

cdef extern from "SDL_image.h":
    ctypedef enum IMG_InitFlags:
        IMG_INIT_JPG
        IMG_INIT_PNG
        IMG_INIT_TIF
        IMG_INIT_WEBP
    cdef int IMG_Init(IMG_InitFlags flags)
    cdef char *IMG_GetError()
    cdef SDL_Surface *IMG_Load(char *file)
    cdef SDL_Surface *IMG_Load_RW(SDL_RWops *src, int freesrc)
    cdef SDL_Surface *IMG_LoadTyped_RW(SDL_RWops *src, int freesrc, char *type)
    cdef int *IMG_SavePNG(SDL_Surface *src, char *file)


class BitmapFontException(Exception):
    """Exception returned by the bitmap font module"""
    pass


cdef class BitmapFontCharacter:
    cdef public int id
    cdef public SDL_Rect rect
    cdef public int xoffset
    cdef public int yoffset
    cdef public int xadvance

    def __cinit__(self, *args, **kw):
        self.id = 0
        self.rect.x = 0
        self.rect.y = 0
        self.rect.w = 0
        self.rect.h = 0
        self.xoffset = 0
        self.yoffset = 0
        self.xadvance = 0


cdef class BitmapFont:
    cdef public str face
    cdef public bint bold
    cdef public bint italic
    cdef public int padding
    cdef public int spacing
    cdef public int outline

    cdef public int line_height
    cdef public int base
    cdef public int scale_w
    cdef public int scale_h

    cdef dict characters
    cdef dict kernings

    cdef SDL_Surface *image

    def __cinit__(self, *args, **kw):
        self.face = ""
        self.bold = False
        self.italic = False
        self.padding = 0
        self.spacing = 0
        self.outline = 0

        self.line_height = 0
        self.base = 0
        self.scale_w = 0
        self.scale_h = 0

        self.characters = dict()
        self.kernings = dict()

        self.image = NULL

    def __init__(self, str image_file, object descriptor):

        # Load the image file
        cdef bytes c_filename = image_file.encode('utf-8')
        cdef SDL_Surface *image = IMG_Load(c_filename)

        if image == NULL:
            raise BitmapFontException("Could not load bitmap font image file")

        self.image = image
        self.scale_w = self.image.w
        self.scale_h = self.image.h

        if isinstance(descriptor, list):
            self._load_descriptor_list(descriptor)
        elif isinstance(descriptor, str):
            self._load_descriptor_file(descriptor)
        else:
            raise BitmapFontException("Illegal value in bitmap font descriptor")

    def __dealloc__(self):
        if self.image != NULL:
            SDL_FreeSurface(self.image)
            self.image = NULL

    def get_image(self):
        image_capsule = pycapsule.PyCapsule_New(self.image, NULL, NULL)
        return image_capsule

    def get_characters(self):
        return self.characters

    def get_kernings(self):
        return self.kernings

    def _load_descriptor_list(self, list descriptor_list):
        """Load the descriptor from the supplied list."""
        cdef int x = 0
        cdef int y = 0
        cdef int row_height = int(self.scale_h / len(descriptor_list))
        cdef int char_width
        cdef int char_id
        self.line_height = row_height
        self.base = row_height

        # Loop over all the rows in the descriptor list
        for row in descriptor_list:
            char_width = int(self.scale_w / len(row))
            x = 0

            # Loop over all characters in the row
            for text_char in row:
                char_id = ord(text_char)

                # Do not add duplicate character definitions (only use the first instance)
                if char_id not in self.characters:

                    # Create character definition
                    character = BitmapFontCharacter()
                    character.id = char_id
                    character.rect.x = x
                    character.rect.y = y
                    character.rect.w = char_width
                    character.rect.h = row_height
                    character.xadvance = char_width

                    self.characters[character.id] = character
                    x += char_width

            y += row_height

    def _load_descriptor_file(self, str descriptor_file):
        """Load the descriptor from the specified file.  Both XML and text
        formats are supported (not binary currently).

        The standard bitmap font descriptor file format can be found at:
        http://www.angelcode.com/products/bmfont/doc/file_format.html
        """
        cdef int char_id
        cdef int first
        cdef int second

        if not path.isfile(descriptor_file):
            raise BitmapFontException('Could not locate the bitmap font descriptor file ' +
                                      descriptor_file)

        # Attempt to parse the file as an XML file
        try:
            xml_tree = ElementTree(file=descriptor_file)
            self._load_descriptor_xml(xml_tree)
        except ParseError:
            pass

        # Assume since the XML parser threw an exception that this is a text file

        # Open the descriptor file
        with open(descriptor_file) as text_file:

            # Loop over all the rows in the descriptor file
            for line in text_file:
                if line.startswith("info"):
                    pass

                elif line.startswith("common"):
                    m = re.search(r"lineHeight=([0-9]{1,5})", line, flags=re.IGNORECASE)
                    if m:
                        self.line_height = int(m.group(1))
                    else:
                        raise BitmapFontException("Bitmap font descriptor file invalid format")

                    m = re.search(r"base=([0-9]{1,5})", line, flags=re.IGNORECASE)
                    if m:
                        self.base = int(m.group(1))
                    else:
                        raise BitmapFontException("Bitmap font descriptor file invalid format")

                elif line.startswith("chars"):
                    pass

                elif line.startswith("char"):
                    m = re.search(r"char\s+id=(?P<id>[0-9]{1,4})\s+x=(?P<x>[0-9]{1,4})\s+y=(?P<y>[0-9]{1,4})"
                                  r"\s+width=(?P<width>[0-9]{1,3})\s+height=(?P<height>[0-9]{1,3})"
                                  r"\s+xoffset=(?P<xoffset>-?[0-9]{1,3})\s+yoffset=(?P<yoffset>-?[0-9]{1,3})"
                                  r"\s+xadvance=(?P<xadvance>[0-9]{1,3})\s+page=(?P<page>[0-9]{1,2})"
                                  r"\s+chnl=(?P<chnl>[0-9]{1,3})", line, flags=re.IGNORECASE)
                    if not m:
                        raise BitmapFontException("Bitmap font descriptor file invalid format")

                    char_id = int(m.group("id"))
                    if char_id > 256:
                        continue

                    # Do not add duplicate character definitions (only use the first instance)
                    if char_id not in self.characters:

                        # Create character definition
                        character = BitmapFontCharacter()
                        character.id = char_id
                        character.rect.x = int(m.group("x"))
                        character.rect.y = int(m.group("y"))
                        character.rect.w = int(m.group("width"))
                        character.rect.h = int(m.group("height"))
                        character.xoffset = int(m.group("xoffset"))
                        character.yoffset = int(m.group("yoffset"))
                        character.xadvance = int(m.group("xadvance"))

                        self.characters[character.id] = character

                elif line.startswith("kernings"):
                    pass

                elif line.startswith("kerning"):
                    m = re.search(r"kerning\s+first=(?P<first>[0-9]{1,3})\s+second=(?P<second>[0-9]{1,3})"
                                  r"\s+amount=(?P<amount>-?[0-9]{1,3})", line, flags=re.IGNORECASE)
                    if not m:
                        raise BitmapFontException("Bitmap font descriptor file invalid format")

                    first = int(m.group("first"))
                    second = int(m.group("second"))

                    if first not in self.kernings:
                        self.kernings[first] = {}

                    if second not in self.kernings[first]:
                        self.kernings[first][second] = int(m.group("amount"))

    def _load_descriptor_xml(self, xml_tree: ElementTree):
        """Loads the descriptor information from the XML tree."""

        cdef int first = 0
        cdef int second = 0
        root = xml_tree.getroot()
        if root is None:
            raise BitmapFontException("Bitmap font descriptor file invalid XML format")

        # info
        info = root.find('info')
        if info is None:
            raise BitmapFontException("Bitmap font descriptor file invalid XML format")

        # common
        common = root.find('common')
        if common is None:
            raise BitmapFontException("Bitmap font descriptor file invalid XML format")

        self.line_height = int(common.attrib["lineHeight"])
        self.base = int(common.attrib["base"])

        # characters
        chars = root.find('chars')
        if chars is None:
            raise BitmapFontException("Bitmap font descriptor file invalid XML format")

        for text_char in chars.findall('char'):
            character = BitmapFontCharacter()
            character.id = int(text_char.attrib["id"])
            character.rect.x = int(text_char.attrib["x"])
            character.rect.y = int(text_char.attrib["y"])
            character.rect.w = int(text_char.attrib["width"])
            character.rect.h = int(text_char.attrib["height"])
            character.xadvance = int(text_char.attrib["xadvance"])
            character.xoffset = int(text_char.attrib["xoffset"])
            character.yoffset = int(text_char.attrib["yoffset"])

            self.characters[character.id] = character

        # kerning
        kernings = root.find('kernings')
        if kernings is not None:
            for kerning in kernings.findall('kerning'):
                first = int(kerning.attrib["first"])
                second = int(kerning.attrib["second"])

                if first not in self.kernings:
                    self.kernings[first] = {}

                self.kernings[first][second] = int(kerning.attrib["amount"])


    def get_descent(self):
        return self.base - self.line_height

    def get_ascent(self):
        return self.base

    def get_extents(self, str text, font_kerning=True):
        """Calculate the actual size of the specified text when rendered."""
        cdef int left_x = 0
        cdef int right_x = 0
        cdef int cursor_x = 0
        cdef int previous_char = -1
        cdef int current_char
        cdef bint use_kerning = font_kerning
        cdef BitmapFontCharacter char_info

        # loop over characters getting character information
        for index, text_char in enumerate(text):
            current_char = ord(text_char)
            if current_char in self.characters:
                char_info = self.characters[current_char]

                # capture the left coordinate of the first character
                if index == 0:
                    left_x = char_info.xoffset

                # apply kerning (if applicable)
                if use_kerning and previous_char in self.kernings and current_char in self.kernings[previous_char]:
                    cursor_x += self.kernings[previous_char][current_char]

                # determine right coordinate of current character
                right_x = cursor_x + char_info.xoffset + char_info.rect.w

                # advance the cursor to measure the next character
                cursor_x += char_info.xadvance

            previous_char = current_char

        return right_x - left_x, self.line_height


cdef class _SurfaceContainer:
    cdef SDL_Surface* surface
    cdef int w, h

    def __cinit__(self, w, h):
        self.surface = NULL
        self.w = w
        self.h = h

    def __init__(self, w, h):
        # XXX check on OSX to see if little endian/big endian make a difference here.
        self.surface = SDL_CreateRGBSurface(0, w, h, 32, 0x000000ff, 0x0000ff00, 0x00ff0000, 0xff000000)
        memset(self.surface.pixels, 0, w * h * 4)

    def __dealloc__(self):
        if self.surface != NULL:
            SDL_FreeSurface(self.surface)
            self.surface = NULL

    def render(self, container, text, x, y):
        """Render the text at the specified location."""
        cdef SDL_Rect cursor_rect
        cdef SDL_Rect source_rect
        cdef SDL_Rect dest_rect
        cdef SDL_Surface *source_image
        cdef dict characters = dict()
        cdef dict kernings = dict()
        cdef bint use_kerning = True
        cdef int previous_char = -1
        cdef int current_char

        asset = container.get_font_asset()
        if asset is None:
            return

        font = asset.bitmap_font
        if font is None:
            return

        source_image = <SDL_Surface*>pycapsule.PyCapsule_GetPointer(font.get_image(), NULL)
        if source_image == NULL:
            return

        """
        How to interpret the values in the font descriptor file:
        line_height is how far the cursor should be moved vertically when moving to the next line. The base value is 
        how far from the top of the cell height the base of the characters in the font should be placed. Characters can 
        extend above or below this base line, which is entirely up to the font design.
        
        The current cursor position determines the starting point for drawing the next character. To get to the next
        cursor position (for the next character) the cursor is moved horizontally in the amount determined by the 
        xadvance value. If kerning pairs are used the cursor should also be moved accordingly. 
        
        The yoffset gives the distance from the top of the cell height to the top of the character. A negative value 
        here would mean that the character extends above the cell height. The xoffset gives the horizontal offset that 
        should be added to the cursor position to find the left position where the character should be drawn. A 
        negative value here would mean that the character slightly overlaps the previous character. Observe that these 
        values shouldn't be used to move the cursor position, only to determine where to actually draw the current
        character.
        
        The x, y, width, and height attributes indicate the precise location and size in the font atlas to copy to the
        current font position.
        """

        cursor_rect.x = x
        cursor_rect.y = y

        characters = font.get_characters()
        kernings = font.get_kernings()
        use_kerning = container.options['font_kerning']

        # Copy each character in the text string from the bitmap font atlas to the output texture/surface
        for index, text_char in enumerate(text):
            current_char = ord(text_char)
            if current_char in characters:
                char_info = characters[current_char]

                # need to ensure the first character is drawn at position 0 (adjust for initial xoffset value)
                if index == 0:
                    cursor_rect.x -= char_info.xoffset

                # apply kerning (if applicable)
                if use_kerning and previous_char in kernings and current_char in kernings[previous_char]:
                    cursor_rect.x += kernings[previous_char][current_char]

                # set rectangles for character image copy and perform image copy operation
                source_rect = char_info.rect
                dest_rect.x = cursor_rect.x + char_info.xoffset
                dest_rect.y = cursor_rect.y + char_info.yoffset
                SDL_BlitSurface(source_image, &source_rect, self.surface, &dest_rect)

                # advance the cursor in preparation for drawing the next character
                cursor_rect.x += char_info.xadvance

            previous_char = current_char

    def get_data(self):
        """Return the bitmap font surface as ImageData (pixels)."""
        cdef int datalen = self.surface.w * self.surface.h * 4
        cdef bytes pixels = (<char *>self.surface.pixels)[:datalen]
        data = ImageData(self.w, self.h, 'rgba', pixels)
        return data



from mpf.core.placeholder_manager import MachinePlaceholder, BasePlaceholderManager


class McPlaceholderManager(BasePlaceholderManager):

    """Manages templates and placeholders for MC."""

    def get_global_parameters(self, name):
        """Return global params."""
        if name == "settings":
            return self.machine.settings
        elif name == "machine":
            return False
        elif self.machine.player:
            if name == "current_player":
                return self.machine.player
            elif name == "players":
                return self.machine.player_list
        return False



"""Test high score mode."""
from collections import OrderedDict
import time

from mpf.tests.MpfFakeGameTestCase import MpfFakeGameTestCase

from mpfmc.tests.MpfSlideTestCase import MpfSlideTestCase

from mpfmc.tests.MpfIntegrationTestCase import MpfIntegrationTestCase


class TestHighScoreMode(MpfIntegrationTestCase, MpfSlideTestCase, MpfFakeGameTestCase):

    def get_config_file(self):
        return 'high_score.yaml'

    def get_machine_path(self):
        return 'integration/machine_files/high_score/'

    def test_empty_name(self):
        self.start_game()
        self.machine.game.player_list[0].score = 10000
        self.machine.game.player_list[0].loops = 100
        self.machine.game.end_game()
        self.advance_time_and_run()
        self.assertTrue(self.machine.modes.high_score.active)
        self.hit_and_release_switch("s_left_flipper")
        self.advance_time_and_run(.1)
        self.assertTextOnTopSlide("PLAYER 1")
        self.assertTextOnTopSlide("GRAND CHAMPION")
        self.assertTextOnTopSlide("END")
        self.hit_and_release_switch("s_start")
        self.advance_time_and_run(.5)
        self.assertTextOnTopSlide("GRAND CHAMPION")
        self.assertTextOnTopSlide("")
        self.advance_time_and_run(5)
        self.assertTextOnTopSlide("LOOP CHAMP")
        self.assertTextOnTopSlide("")

    def test_no_high_scores(self):
        self.start_game()
        self.machine.game.player_list[0].score = 10000
        self.machine.game.player_list[0].loops = 100
        self.machine.game.end_game()
        self.advance_time_and_run()
        self.assertTrue(self.machine.modes.high_score.active)

        self.assertTextOnTopSlide("PLAYER 1")
        self.assertTextOnTopSlide("GRAND CHAMPION")
        self.assertTextOnTopSlide("A")

        for i in range(9):
            self.hit_and_release_switch("s_right_flipper")
        self.advance_time_and_run()
        self.assertTextOnTopSlide("J")

        self.hit_and_release_switch("s_start")
        for i in range(9):
            self.hit_and_release_switch("s_left_flipper")
        self.advance_time_and_run()
        self.assertTextOnTopSlide("A")

        self.hit_and_release_switch("s_start")
        self.advance_time_and_run()

        self.assertTextOnTopSlide("JA")

        self.hit_and_release_switch("s_right_flipper")
        self.hit_and_release_switch("s_start")
        self.advance_time_and_run()

        self.assertTextOnTopSlide("JAB")

        self.hit_and_release_switch("s_start")
        self.advance_time_and_run(.5)
        self.assertTextOnTopSlide("GRAND CHAMPION")
        self.assertTextOnTopSlide("JAB")
        self.advance_time_and_run(5)
        self.assertTextOnTopSlide("LOOP CHAMP")
        self.assertTextOnTopSlide("JAB")
        self.advance_time_and_run(5)

        self.assertFalse(self.machine.modes.high_score.active)



from mpfmc.uix.slide import Slide
from mpfmc.tests.MpfMcTestCase import MpfMcTestCase


class TestSlides(MpfMcTestCase):
    def get_machine_path(self):
        return 'tests/machine_files/slide'

    def get_config_file(self):
        return 'test_slides.yaml'

    def test_slide(self):
        slide = Slide(mc=self.mc, name='slide1', priority=100)

        self.assertEqual(slide.mc, self.mc)
        self.assertEqual(slide.name, 'slide1')
        self.assertEqual(slide.priority, 100)

    def test_mc_slide_processor(self):
        self.assertIn('slide1', self.mc.machine_config['slides'])
        self.assertIn('slide2', self.mc.machine_config['slides'])
        self.assertIn('slide3', self.mc.machine_config['slides'])
        self.assertIn('slide4', self.mc.machine_config['slides'])
        self.assertIn('slide5', self.mc.machine_config['slides'])

    def test_key_by_name(self):
        slide = Slide(mc=self.mc,
                      name='slide1',
                      config=self.mc.slides['slide1'],
                      key='mode1',
                      priority=0)

        self.assertEqual(slide.key, 'mode1')

    def test_priority_passed(self):
        slide = Slide(mc=self.mc,
                      name='slide1',
                      config=self.mc.slides['slide1'],
                      key='mode1',
                      priority=123)

        self.assertEqual(slide.priority, 123)

    def test_widgets_from_config(self):
        slide = Slide(mc=self.mc,
                      name='slide1',
                      config=self.mc.slides['slide1'],
                      key='mode1',
                      priority=0)

        widget_tree = list()
        for s in slide.children:
            widget_tree.append(s.widget)

        # Widgets are drawn in order from the end of the children list to the beginning
        # so the order should be the same order as appears in the config file.
        self.assertEqual(widget_tree[0].text, 'SLIDE TEST 1-1')
        self.assertEqual(widget_tree[1].text, 'SLIDE TEST 1-2')
        self.assertEqual(widget_tree[2].text, 'SLIDE TEST 1-3')

    def test_widgets_z_order_from_config(self):
        slide = Slide(mc=self.mc,
                      name='slide7',
                      config=self.mc.slides['slide7'],
                      key='mode1',
                      priority=0)

        widget_tree = list()
        for s in slide.children:
            widget_tree.append(s.widget)

        # Widgets are drawn in order from the end of the children list to the beginning.
        # The resulting child widget order should be the from highest z-order value in
        # the beginning of the list to lowest z-order value.  Ties should be broken by
        # the order the widgets appear in the config file.
        self.assertEqual(widget_tree[0].text, 'TEST Z-ORDER 100')
        self.assertEqual(widget_tree[1].text, 'TEST Z-ORDER 50-1')
        self.assertEqual(widget_tree[2].text, 'TEST Z-ORDER 50-2')
        self.assertEqual(widget_tree[3].text, 'TEST Z-ORDER 0')

    def test_slides_from_config(self):
        self.assertIn('slide1', self.mc.slides)
        self.assertIn('slide2', self.mc.slides)
        self.assertIn('slide3', self.mc.slides)
        self.assertIn('slide4', self.mc.slides)
        self.assertIn('transition', self.mc.slides['slide4'])
        self.assertIn('slide5', self.mc.slides)
        self.assertIn('transition', self.mc.slides['slide5'])

        # make sure it also works by attribute
        self.assertIsNotNone(self.mc.slides.slide1)

    def test_quick_actions(self):
        # There was a bug where immediately adding and then removing a slide
        # would crash because the kivy event loop didn't get a chance to run.

        self.mc.targets['default'].show_slide('slide1')
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'slide1')
        self.mc.targets['default'].remove_slide('slide1')

    def test_background_color_and_opacity(self):
        slide = Slide(mc=self.mc,
                      name='slide6',
                      config=self.mc.slides['slide6'])
        self.advance_time()

        self.assertEqual(0.5, slide.canvas.opacity)

    def test_remove_non_existent_slide(self):
        self.assertFalse(self.mc.targets['default'].remove_slide('hello'))
        self.advance_time()

    def test_two_slides_in_one_tick(self):
        self.mock_event('slide_slide1_active')
        self.mock_event('slide_slide2_active')

        self.mc.targets['default'].show_slide('slide1')
        self.mc.targets['default'].show_slide('slide2')
        self.advance_time()

        self.assertEventNotCalled('slide_slide1_active')
        self.assertEventCalled('slide_slide2_active', 1)



from mpf.core.assets import Asset


# pylint: disable-msg=abstract-method
class McAsset(Asset):

    """Baseclass for all assets in mc."""

    __slots__ = ['__weakref__']

    def __init__(self, machine, name, file, config):
        """Track this asset for potential leaks."""
        super().__init__(machine, name, file, config)
        machine.track_leak_reference(self)



"""Contains the playlist player class"""

from copy import deepcopy
from mpf.core.config_validator import ConfigValidator
from mpfmc.core.mc_config_player import McConfigPlayer


class McPlaylistPlayer(McConfigPlayer):
    """Base class for the Playlist Player that runs on the mpf-mc side of things.
    It receives all of its instructions via BCP from a MpfPlaylistPlayer instance
    running as part of MPF.

    The playlist_player: section of a config file (either the machine-wide or
    a mode-specific config) looks like this:

    playlist_player:
        <event_name>:
            <playlist_track_name>:
                <playlist_settings>: ...

    Express config is not supported in the playlist player.

    To control other settings (such as track, action, etc.), enter the playlist name on
    the next line and the settings indented under it, like this:

    playlist_player:
        some_event:
            playlist_track_name:
                action: advance

Here are several various examples:

    playlist_player:
        some_event:
            playlist:
                playlist: attract_mode_music
                volume: 0.65

        some_event2:
            playlist:
                action: stop
                fade-out: 2s

    """

    config_file_section = 'playlist_player'
    show_section = 'playlists'
    machine_collection_name = 'playlists'

    # pylint: disable=invalid-name
    def play(self, settings, context, calling_context, priority=0, **kwargs):
        """Plays a validated section from a playlist_player: section of a
        config file or the playlists: section of a show.

        The config must be validated.
        """
        del calling_context
        settings = deepcopy(settings)

        self.machine.log.debug("PlaylistPlayer: Play called with settings=%s", settings)

        for track_name, player_settings in settings.items():
            player_settings.update(kwargs)

            playlist_controller = self.machine.sound_system.audio_interface.get_playlist_controller(track_name)

            # Determine action to perform
            if player_settings['action'].lower() == 'play':
                if player_settings['playlist'] not in self.machine.playlists:
                    self.machine.log.error("PlaylistPlayer: The specified playlist "
                                           "does not exist ('{}').".format(player_settings['playlist']))
                    return
                try:
                    playlist_name = player_settings['playlist']
                    del player_settings['playlist']
                    playlist_controller.play(playlist_name, context, player_settings)
                except Exception as ex:
                    raise Exception(ex)

            elif player_settings['action'].lower() == 'stop':
                player_settings.setdefault('fade_out', None)
                playlist_controller.stop()

            elif player_settings['action'].lower() == 'advance':
                playlist_controller.advance()

            elif player_settings['action'].lower() == 'set_repeat':
                player_settings.setdefault('repeat', True)
                playlist_controller.set_repeat(player_settings['repeat'])

            else:
                self.machine.log.error("SoundLoopPlayer: The specified action "
                                       "is not valid ('{}').".format(player_settings['action']))

    def get_express_config(self, value):
        """There is no express config for the playlist player."""
        del value
        raise AssertionError("Playlist Player does not support express config")

    # pylint: disable=too-many-branches
    def validate_config(self, config):
        """Validates the playlist_player: section of a config file (either a
        machine-wide config or a mode config).

        Args:
            config: A dict of the contents of the playlist_player section
            from the config file. It's assumed that keys are event names, and
            values are settings for what the playlist_player should do when
            that event is posted.

        Returns: A dict a validated entries.

        This method overrides the base method since the playlist_player has
        unique options.

        """
        validated_config = dict()

        # No need to validate if sound system is not enabled, just return empty dict
        if self.machine.sound_system is None or self.machine.sound_system.audio_interface is None:
            return validated_config

        for event, settings in config.items():
            validated_config[event] = dict()

            for track_name, player_settings in settings.items():

                # Validate the specified track name is a sound_loop track
                playlist_controller = self.machine.sound_system.audio_interface.get_playlist_controller(track_name)
                if playlist_controller is None or \
                        playlist_controller.track is None or \
                        playlist_controller.track.type != "standard":
                    raise ValueError("PlaylistPlayer: An invalid audio track '{}' is specified for event '{}' "
                                     "(only playlist audio tracks are supported).".format(track_name, event))

                validated_config[event].update(self._validate_config_item(track_name, player_settings))

        return validated_config

    def _validate_config_item(self, device, device_settings):
        """Validates the config when in a show or in a player"""

        # event contains the event name that triggers the playlist_player action
        # Validate the settings against the config spec

        # First validate the action item (since it will be used to validate the rest
        # of the config)
        if 'action' in device_settings:
            device_settings['action'] = self.machine.config_validator.validate_config_item(
                self.machine.config_validator.config_spec['playlist_player']['action'],
                'playlist_player:{}'.format(device),
                device_settings['action'])
        else:
            device_settings['action'] = self.machine.config_validator.validate_config_item(
                self.machine.config_validator.config_spec['playlist_player']['action'],
                'playlist_player:{}'.format(device))

        validated_settings = self.machine.config_validator.validate_config(
            'playlist_player_actions:{}'.format(device_settings['action']).lower(),
            device_settings)

        # Remove any items from the settings that were not explicitly provided in the
        # playlist_player config section (only want to override settings explicitly
        # and not with any default values).  The default values for these items are not
        # legal values and therefore we know the user did not provide them.
        if 'crossfade_mode' in validated_settings and validated_settings['crossfade_mode'] == 'use_playlist_setting':
            del validated_settings['crossfade_mode']
        if 'crossfade_time' in validated_settings and validated_settings['crossfade_time'] is None:
            del validated_settings['crossfade_time']
        if 'volume' in validated_settings and validated_settings['volume'] is None:
            del validated_settings['volume']
        if 'fade_in' in validated_settings and validated_settings['fade_in'] is None:
            del validated_settings['fade_in']
        if 'shuffle' in validated_settings and validated_settings['shuffle'] is None:
            del validated_settings['shuffle']
        if 'repeat' in validated_settings and validated_settings['repeat'] is None:
            del validated_settings['repeat']
        if 'scope' in validated_settings and validated_settings['scope'] == 'use_playlist_setting':
            del validated_settings['scope']
        if 'events_when_played' in validated_settings and len(validated_settings['events_when_played']) == 1 and \
                validated_settings['events_when_played'][0] == 'use_playlist_setting':
            del validated_settings['events_when_played']
        if 'events_when_stopped' in validated_settings and len(validated_settings['events_when_stopped']) == 1 and \
                validated_settings['events_when_stopped'][0] == 'use_playlist_setting':
            del validated_settings['events_when_stopped']
        if 'events_when_looping' in validated_settings and len(validated_settings['events_when_looping']) == 1 and \
                validated_settings['events_when_looping'][0] == 'use_playlist_setting':
            del validated_settings['events_when_looping']
        if 'events_when_sound_changed' in validated_settings and len(
                validated_settings['events_when_sound_changed']) == 1 and \
                validated_settings['events_when_sound_changed'][0] == 'use_playlist_setting':
            del validated_settings['events_when_sound_changed']
        if 'events_when_sound_stopped' in validated_settings and len(
                validated_settings['events_when_sound_stopped']) == 1 and \
                validated_settings['events_when_sound_stopped'][0] == 'use_playlist_setting':
            del validated_settings['events_when_sound_stopped']

        validated_dict = dict()
        validated_dict[device] = validated_settings
        return validated_dict

    def clear_context(self, context):
        """Stop all sounds from this context."""
        self.machine.log.debug("PlaylistPlayer: Clearing context - "
                               "stopping any active playlists started from this context")

        for name in self.machine.sound_system.audio_interface.get_playlist_controller_names():
            playlist_controller = self.machine.sound_system.audio_interface.get_playlist_controller(name)
            if playlist_controller:
                playlist_controller.clear_context(context)


mc_player_cls = McPlaylistPlayer



"""Contains the Display base class, which is a logical display in the mpf-mc."""
from typing import List, Union, Optional
from math import floor

from kivy.uix.floatlayout import FloatLayout

from kivy.clock import Clock
from kivy.uix.screenmanager import (ScreenManager, NoTransition,
                                    SlideTransition, SwapTransition,
                                    FadeTransition, WipeTransition,
                                    FallOutTransition, RiseInTransition,
                                    ScreenManagerException)
from kivy.uix.widget import Widget as KivyWidget, WidgetException as KivyWidgetException
from kivy.uix.scatter import Scatter
from kivy.graphics import (
    Translate, Fbo, ClearColor, ClearBuffers, Scale)
from kivy.properties import ObjectProperty

from mpfmc.uix.widget import WidgetContainer, Widget
from mpfmc.uix.slide import Slide


MYPY = False
if MYPY:   # pragma: no cover
    from mpfmc.core.mc import MpfMc


transition_map = dict(none=NoTransition,
                      slide=SlideTransition,
                      swap=SwapTransition,
                      fade=FadeTransition,
                      wipe=WipeTransition,
                      fall_out=FallOutTransition,
                      rise_in=RiseInTransition)


# pylint: disable-msg=too-many-instance-attributes
class Display(ScreenManager):

    """A display which can be used to show slides."""

    displays_to_initialize = 0

    texture = ObjectProperty(None, allownone=True)

    @staticmethod
    def create_default_display(mc: "MpfMc") -> None:
        """Create default display."""
        Display(mc, 'default', width=800, height=600)

    def __init__(self, mc: "MpfMc", name: str, **kwargs) -> None:
        """Initialise Display."""
        self.mc = mc
        self.name = name
        self.config = kwargs
        self._ready = False
        self.tags = []
        self.display = self
        self.parents = []
        self.mc.track_leak_reference(self)

        Display.displays_to_initialize += 1

        self.native_size = (self.config['width'], self.config['height'])
        self.size_hint = (None, None)
        self.size = self.native_size

        self.transition = NoTransition()

        self._blank_slide_name = '{}_blank'.format(self.name)

        super().__init__()

        # It is possible that the current slide changes more than one time during a single clock
        # frame. Sending multiple slide active events is not desired in this situation. This can
        # easily be solved using Kivy clock events. The clock event will only be triggered once
        # per clock frame no matter how many times it is called.
        self._current_slide_changed = Clock.create_trigger(self._post_active_slide_event, -1)

        # Need to create a widget that will be the parent of the slide manager.  This is
        # necessary to allow widgets with negative z values to remain in the display while
        # slides are changed.
        self.container = FloatLayout(size_hint=(None, None), size=self.size)
        self.container.z = 0
        self.container.add_widget(self)

        self._display_created()

    def __repr__(self):
        return '<Display name={}{}, current slide={}, total slides={}>'.format(
            self.name, self.size, self.current_slide_name, len(self.slides))

    def get_frame_data(self, *args):
        """Return the content of this display as buffer.

        @see: widget.export_to_png
        """
        del args

        fbo = Fbo(size=self._slide_manager_parent.size, with_stencilbuffer=True)

        with fbo:
            ClearColor(0, 0, 0, 1)
            ClearBuffers()
            Scale(1, -1, 1)
            Translate(-self.x,
                      -self.y - self.height, 0)

        fbo.add(self.canvas)
        fbo.draw()
        data = fbo.texture.pixels
        fbo.remove(self.canvas)

        return data

    @property
    def ready(self):
        """Return true if display is ready."""
        return self._ready

    @property
    def parent_widgets(self) -> List["WidgetContainer"]:
        """The list of all widgets owned by the display parent."""
        return [x for x in self.container.children if x != self]

    def has_parent(self) -> bool:
        """Returns whether or not the display has a parent."""
        return bool(self.container.parent is not None)

    def _display_created(self, *args) -> None:
        """Callback after this display is created."""
        del args
        # There's a race condition since mpf-mc will continue while the display
        # gets setup. So we need to wait to know that the display is done.
        # Easiest way to do that is to check to see if the display is the right
        # size, and when it is, we move on.
        if (self.size[0] != self.native_size[0] or
                self.size[1] != self.native_size[1]):
            self.size = self.native_size
            Clock.schedule_once(self._display_created, 0)
            return
        # Add this display to the list of all available displays
        self.mc.displays[self.name] = self

        # If this display is configured as the default, set it.
        # If this display would overwrite an existing default, raise an AssertionError.
        try:
            if self.config['default']:
                if 'default' not in self.mc.targets:
                    self.mc.targets['default'] = self
                else:
                    raise AssertionError('Multiple displays have been set as the default. Please choose a single \
                        display to default to (\"{}\" is currently set as the default).'.format(
                        self.mc.targets['default'].name))
        except KeyError:
            pass

        # Initialization is just about done, schedule callback to finish
        Clock.schedule_once(self._init_done, 0)

    def _init_done(self, *args) -> None:
        """Callback after this display has been initialized."""
        del args
        self.mc.post_mc_native_event('display_{}_initialized'.format(self.name))
        '''event: display_(name)_initialized
        config_section: displays
        class_label: display

        desc: The display called (name) has been initialized. This event is
        generated in the MC, so it won't be sent to MPF if the MC is started up
        and ready first.

        This event is part of the MPF-MC boot process and is not particularly
        useful for game developers. If you want to show a "boot" slide as
        early as possible, use the *mc_ready* event.
        '''

        Display.displays_to_initialize -= 1

        # Callback function to set this display to ready state once all displays
        # have been initialized
        self.mc.events.add_handler('displays_initialized', self._finalize_setup, priority=10000)

        if not Display.displays_to_initialize:
            Clock.schedule_once(self._displays_initialized)

    def _displays_initialized(self, *args) -> None:
        """Callback after all displays have been initialized."""
        del args

        # Determine the 'default' display
        if len(self.mc.displays) == 1:
            self.mc.targets['default'] = \
                [x for x in self.mc.displays.values()][0]

        elif 'default' not in self.mc.targets:
            for target in ('window', 'dmd'):
                if target in self.mc.displays:
                    self.mc.targets['default'] = self.mc.displays[target]
                    break

            if 'default' not in self.mc.targets:
                self.mc.targets['default'] = self.mc.displays[
                    (sorted(self.mc.displays.keys()))[0]]

        self.mc.log.info("Display: Setting '%s' as default display", self.mc.targets['default'].name)

        self.mc.displays_initialized()

    def _finalize_setup(self, **kwargs) -> None:
        """Callback function after all displays have been initialized.  This
        method finalizes the display setup and gets it ready to use as a
        target. The 'display_{}_ready' event is posted once it is ready.
        """
        del kwargs

        # This display is now a valid target so add it to the list
        self.mc.targets[self.name] = self

        # Create a blank slide for this display. Why?
        # 1. sometimes people try to play widgets with no slide. This makes
        # that work.
        # 2. the first slide that's created and added to this frame will be
        # automatically shown, which we don't want. Also we want to ensure that
        # our slide event will be called which only happens when this slide is
        # switched to, rather than automatically added.
        self.create_blank_slide()

        self._ready = True

        self.mc.post_mc_native_event('display_{}_ready'.format(self.name))
        '''event: display_(name)_ready
        config_section: displays
        class_label: display

        desc: The display target called (name) is now ready and available to
        show slides.
        This event is useful with display widgets where you want to add
        a display to an existing slide which shows some content, but you
        need to make sure the display exists before showing a slide.
        So if you have a display called "overlay", then you can add it to
        a slide however you want, and when it's added, the event
        "display_overlay_ready" will be posted, and then you can use that event
        in your slide_player to trigger the first slide you want to show.
        Note that this event is posted by MPF-MC and will not exist on the MPF
        side. So you can use this event for slide_player, widget_player, etc.,
        but not to start shows or other things controlled by MPF.'''

    @property
    def current_slide(self) -> "Slide":
        """Returns the Slide object of the current slide."""
        return self.current_screen

    @current_slide.setter
    def current_slide(self, value: Union[str, "Slide"]):
        """Set the current slide.
        You can set it to a Slide object or a string of the slide name.
        """
        if isinstance(value, Slide):
            self._set_current_slide(value)
        elif isinstance(value, str):
            self._set_current_slide_name(value)

    @property
    def current_slide_name(self) -> str:
        """Returns the string name of the current slide."""
        return self.current

    @current_slide_name.setter
    def current_slide_name(self, value: str):
        """Sets the current slide based on the string name of the slide you
        want to be shown."""
        self._set_current_slide_name(value)

    @property
    def slides(self) -> List["Slide"]:
        """Return list of slide objects of all the active slides for this slide frame."""
        return self.screens

    def create_blank_slide(self) -> "Slide":
        """Creates the blank slide for this display."""
        return self.add_slide(self._blank_slide_name)

    def get_slide(self, name: str) -> "Slide":
        """Return the Slide associated with the name or raise a
        :class:`ScreenManagerException` if not found."""
        return self.get_screen(name)

    # pylint: disable-msg=too-many-arguments
    def add_slide(self, name: str, config: Optional[dict] = None, priority: int = 0,
                  key: Optional[str] = None, play_kwargs: Optional[dict] = None) -> "Slide":
        """Add a slide to this display.

        Add a slide to the list of slides managed by the display (or returns the existing
        slide with the specified name if it already exists).  This method just adds the
        slide.  It does not display it.

        Args:
            name: The slide name.
            config: The slide config.
            priority: The slide priority.
            key: Optional key.
            play_kwargs: Additional play kwargs.

        Returns:
            The Slide object.
        """
        # See if slide already exists.  If so, return it
        if self.has_screen(name):
            return self.get_screen(name)

        # Slide() creates a new slide and adds it to this screen manager (display)
        return Slide(mc=self.mc, name=name, target=self.name,
                     config=config, key=key, priority=priority,
                     play_kwargs=play_kwargs)

    # pylint: disable-msg=too-many-arguments
    def show_slide(self, slide_name: str, transition: Optional[str] = None,
                   key: Optional[str] = None, force: bool = False, priority: int = 0,
                   show: Optional[bool] = True, expire: Optional[float] = None,
                   play_kwargs: Optional[dict] = None, **kwargs) -> bool:
        """
        Request to show the specified slide. Many of the slide parameters may be overridden
        using the arguments for this function.

        Args:
            slide_name: The name of the slide.
            transition: The slide transition (overrides any stored in the slide).
            key: The slide key.
            force: When true, the slide will be displayed regardless of the priority of the
                current slide.
            priority: The priority of the slide to show.
            show: Whether or not to actually show the slide.
            expire: Expiration time (in seconds) after which the slide will be automatically
                removed (overrides value stored in the slide).
            play_kwargs: Kwargs related to playing/displaying the slide.
            **kwargs: Additional kwargs (will override settings in the play_kwargs parameter).

        Returns:
            True is the slide will be shown, False otherwise.
        """
        # TODO: Is the show parameter really needed?  Why call show_slide and not show the slide?
        if not play_kwargs:
            play_kwargs = kwargs
        else:
            play_kwargs.update(kwargs)

        if self.has_screen(slide_name):
            slide = self.get_screen(slide_name)
        else:
            try:
                slide_config = self.mc.slides[slide_name]
            except KeyError:
                raise AssertionError("Slide {} not found".format(slide_name))
            slide = self.add_slide(name=slide_name,
                                   config=slide_config,
                                   priority=priority,
                                   key=key,
                                   play_kwargs=play_kwargs)

        # update the widgets with whatever kwargs came through here
        if play_kwargs:
            for widget in slide.walk():
                try:
                    widget.update_kwargs(**play_kwargs)
                except AttributeError:
                    pass

        if not transition:
            try:  # anon slides are in the collection
                transition = self.mc.slides[slide_name]['transition']
            except KeyError:
                pass

        # If there's an expire kwarg, that takes priority over slide's expire
        if expire:
            slide.schedule_removal(expire)
        elif slide.expire:
            slide.schedule_removal(slide.expire)

        if (slide.priority >= self.current_slide.priority and show) or force:
            # We need to show this slide

            # Have to set a transition even if there's not one because we have
            # to remove whatever transition was last used
            self.transition.stop()
            self.transition = self.mc.transition_manager.get_transition(transition)

            self._set_current_slide(slide)
            return True

        else:
            # Not showing this slide
            return False

    # pylint: disable-msg=too-many-arguments
    def add_and_show_slide(self, widgets: Optional[dict] = None,
                           slide_name: Optional[str] = None,
                           transition: Optional[str] = None, priority: int = 0,
                           key: Optional[str] = None, force: bool = False,
                           background_color=None,
                           expire: Optional[float] = None, play_kwargs=None,
                           **kwargs) -> bool:
        """Create and show the slide.

        If a slide with this name already exists, it will be replaced.

        Args:
            widgets: An optional dictionary of widgets to add to the slide.
            slide_name: The name of the slide.
            transition: The slide transition (overrides any stored in the slide).
            force: When true, the slide will be displayed regardless of the priority of the
                current slide.
            key: The slide key.
            priority: The priority of the slide to show.
            expire: Expiration time (in seconds) after which the slide will be automatically
                removed (overrides value stored in the slide).
            play_kwargs: Kwargs related to playing/displaying the slide.
            **kwargs: Additional kwargs (will override settings in the play_kwargs parameter).

        Returns:
            True is the slide will be shown, False otherwise.
        """
        if not play_kwargs:
            play_kwargs = kwargs
        else:
            play_kwargs.update(kwargs)

        slide_obj = self.add_slide(name=slide_name,
                                   config=dict(widgets=widgets, background_color=background_color),
                                   priority=priority, key=key)

        return self.show_slide(slide_name=slide_obj.name, transition=transition,
                               priority=priority, force=force, key=key,
                               expire=expire, play_kwargs=play_kwargs)

    def remove_slide(self, slide: Union["Slide", str],
                     transition_config: Optional[dict] = None) -> bool:
        """Remove a slide from the display.

        Args:
            slide: The slide to remove (can be name string or Slide object)
            transition_config: Optional dictionary containing the transition configuration
                to use while removing the slide (overrides slide setting).

        Returns:
            True if the slide is scheduled to be removed, False otherwise

        Notes:
            You can't remove the automatically generated blank slide, so if you try it will
            raise an exception.
        """
        # TODO:
        # Warning, if you just created a slide, you have to wait at least on
        # tick before removing it. Can we prevent that? What if someone tilts
        # at the exact perfect instant when a mode was starting or something?

        # maybe we make sure to run a Kivy tick between bcp reads or something?
        try:
            slide = self.get_slide(slide)
        except ScreenManagerException:  # no slide by that name
            if not isinstance(slide, Slide):
                return False

        # Do not allow the blank slide to be removed
        if slide.name == self._blank_slide_name:
            return False

        slide.prepare_for_removal()

        self.mc.active_slides.pop(slide.name, None)

        # If the current slide is the active one, find the next highest
        # priority one to show instead.
        if self.current_slide == slide:
            new_slide = self._get_next_highest_priority_slide(slide)
            if self.transition:
                self.transition.stop()

            if transition_config:
                self.transition = self.mc.transition_manager.get_transition(
                    transition_config)
            elif self.current_slide.transition_out:
                self.transition = self.mc.transition_manager.get_transition(
                    self.current_slide.transition_out)
            else:
                self.transition = NoTransition()

            self.transition.bind(on_complete=self._remove_transition)
        else:
            new_slide = None

        # Set the new slide first, so we can transition out of the old before removing
        if new_slide:
            self._set_current_slide(new_slide)
        try:
            self.remove_widget(slide)
        except ScreenManagerException:
            return False

        return True

    def _remove_transition(self, transition):
        """Remove transition if done."""
        if self.transition == transition:
            self.transition = NoTransition()

    def _set_current_slide(self, slide: "Slide"):
        # slide frame requires at least one slide, so if you try to set current
        # to None, it will create a new slide called '<display name>_blank' at
        # priority 0 and show that one

        # I think there's a bug in Kivy 1.9.1. According to the docs, you
        # should be able to set self.current to a screen name. But if that
        # screen is already managed by this screen manager, it will raise
        # an exception, and the source is way deep in their code and not
        # easy to fix by subclassing. So this is sort of a hack that looks
        # for that exception, and if it sees it, it just removes and
        # re-adds the screen.
        if not slide:
            slide = self.create_blank_slide()

        if self.current == slide.name:
            return

        try:
            self.current = slide.name
        except KivyWidgetException:
            self.remove_widget(slide)
            self.add_widget(slide)
            self.current = slide.name

        # Post the event via callback at the end of the frame in case more than
        # one slide was set in this frame, so we only want to post the event
        # for the slide that actually became active.  The Kivy clock event will
        # only call the associated callback once per frame when triggered no
        # matter how many times it is called.
        self._current_slide_changed()

    def _set_current_slide_name(self, slide_name):
        try:
            self._set_current_slide(self.get_screen(slide_name))
        except ScreenManagerException:
            raise ValueError('Cannot set current slide to "{}" as there is '
                             'no slide in this slide_frame with that '
                             'name'.format(slide_name))

    def _get_next_highest_priority_slide(self, slide: "Slide") -> "Slide":
        """Return the slide with the next highest priority."""
        new_slide = None

        for s in self.slides:
            if s == slide:
                continue
            elif not new_slide:
                new_slide = s
            elif s.priority > new_slide.priority:
                new_slide = s
            elif (s.priority == new_slide.priority and
                    s.creation_order > new_slide.creation_order):
                new_slide = s

        return new_slide

    def add_widget_to_current_slide(self, widget: "KivyWidget"):
        """Adds the widget to the current slide."""
        self.current_slide.add_widget(widget)

    def add_widgets_to_current_slide(self, widgets: List["KivyWidget"]):
        """Adds a list of widgets to the current slide."""
        for w in widgets:
            self.add_widget_to_current_slide(w)

    def remove_widgets_by_key(self, key: str) -> None:
        """Removes all widgets with the specified key."""
        for widget in self.find_widgets_by_key(key):
            widget.prepare_for_removal()
            if isinstance(widget, Widget) and widget.container and widget.container.parent:
                widget.container.parent.remove_widget(widget.container)
            elif widget.parent:
                widget.parent.remove_widget(widget)

    def find_widgets_by_key(self, key: str) -> List["KivyWidget"]:
        """Retrieves a list of all widgets with the specified key value."""
        widgets = []

        # First find all matching widgets owned by the slide parent
        for child in self.parent_widgets:
            widgets.extend([x for x in child.walk(restrict=True, loopback=False)
                            if hasattr(x, "key") and x.key == key])

        # Finally find all matching widgets owned by each slide
        for slide in self.slides:
            widgets.extend(slide.find_widgets_by_key(key))

        return widgets

    def _post_active_slide_event(self, dt) -> None:
        """Posts an event that a new slide is now active."""
        del dt

        self.mc.post_mc_native_event('slide_{}_active'.format(self.current_slide_name))
        """event: slide_(name)_active
        config_section: slides
        class_label: slide

        desc: A slide called (name) has just become active, meaning that
        it's now showing as the current slide.
        This is useful for things like the widget_player where you want to
        target a widget for a specific slide, but you can only do so if
        that slide exists.
        Slide names do not take into account what display they're playing on,
        so be sure to create machine-wide unique names when you're naming
        your slides.
        """


class DisplayOutput(Scatter):

    """Show a display as a widget."""

    def __init__(self, parent: "KivyWidget", display: "Display", **kwargs):
        kwargs.setdefault('do_scale', False)
        kwargs.setdefault('do_translation', False)
        kwargs.setdefault('do_rotation', False)

        super().__init__(**kwargs)

        self.key = None

        # It is important that the content of this display output does not contain any
        # circular references to the same display (cannot do a recursive
        # picture-in-picture or Kivy will crash).  Detect and prevent that situation.

        # Rather than adding the display as a child of this widget, we will simply
        # add the display's canvas to this widget's canvas.  This allows the display
        # to essentially have multiple parents. The canvas contains all the
        # instructions needed to draw the widgets.
        self.display = display

        parent.bind(size=self.on_parent_resize)
        self._fit_to_parent()

    #
    # Tree management
    #
    def add_display_source(self, widget):
        """Add a new widget as a child of this widget.

        :Parameters:
            `widget`: :class:`Widget`
                Widget to add to our list of children.
        """
        if not isinstance(widget, Display):
            raise KivyWidgetException(
                'add_widget_multi_parent() can be used only with instances'
                ' of the Display class.')

        widget = widget.__self__
        if widget is self:
            raise KivyWidgetException(
                'Widget instances cannot be added to themselves.')

        widget.parent = self
        widget.parents.append(self)

        canvas = self.canvas
        canvas.add(widget.container.canvas)

    def remove_display_source(self, widget):
        """Remove a display."""
        if not isinstance(widget, Display):
            raise KivyWidgetException(
                'remove_display_source() can be used only with instances'
                ' of the Display class.')
        widget.parents.remove(self)
        widget.parent = None
        self.canvas.remove(widget.container.canvas)

    def __repr__(self) -> str:  # pragma: no cover
        try:
            return '<DisplayOutput size={}, pos={}, source={}>'.format(
                self.size, self.pos, self.display.name)
        except AttributeError:
            return '<DisplayOutput size={}, source=(none)>'.format(self.size)

    def on_parent_resize(self, *args):
        """Fit to parent on resize."""
        del args
        self._fit_to_parent()

    def _fit_to_parent(self, *args):
        """Center and scale display output in parent display widget"""
        del args
        if self.parent:
            self.scale = min(self.parent.width / float(self.display.width),
                             self.parent.height / float(self.display.height))
            self.width = floor(self.scale * self.display.width)
            self.height = floor(self.scale * self.display.height)
            self.x = (self.parent.width - self.width) // 2
            self.y = (self.parent.height - self.height) // 2



"""Contains the sound config player class"""

from copy import deepcopy
from mpf.core.config_validator import ConfigValidator
from mpfmc.core.mc_config_player import McConfigPlayer


class McSoundPlayer(McConfigPlayer):
    """Base class for the Sound Player that runs on the mpf-mc side of things.
    It receives all of its instructions via BCP from a MpfSoundPlayer instance
    running as part of MPF.

    The sound_player: section of a config file (either the machine-wide or
    a mode-specific config) looks like this:

    sound_player:
        <event_name>:
            <sound_name>:
                <sound_settings>: ...

    The express config just puts a sound_name next to an event.

    sound_player:
        some_event: sound_name_to_play

    If you want to control other settings (such as track, priority, volume,
    loops, etc.), enter the sound name on the next line and the settings
    indented under it, like this:

    sound_player:
        some_event:
            sound_name_to_play:
                volume: 0.35
                max_queue_time: 1 sec

Here are several various examples:

    sound_player:
        some_event:
            sound1:
                volume: 0.65

        some_event2:
            sound2:
                volume: -4.5 db
                priority: 100
                max_queue_time: 500 ms

        some_event3: sound3

    """
    config_file_section = 'sound_player'
    show_section = 'sounds'
    machine_collection_name = 'sounds'

    # pylint: disable=invalid-name
    def play(self, settings, context, calling_context, priority=0, **kwargs):
        """Plays a validated sounds: section from a sound_player: section of a
        config file or the sounds: section of a show.

        The config must be validated. Validated config looks like this:

        <sound_name>:
            <settings>: ...

        <settings> can be:

        action:
        priority:
        volume:
        loops:
        max_queue_time:

        Notes:
            Ducking settings and markers cannot currently be specified/overridden in the
            sound_player (they must be specified in the sounds section of a config file).

        """
        del calling_context
        settings = deepcopy(settings)

        if 'sounds' in settings:
            settings = settings['sounds']

        for sound_name, s in settings.items():

            try:
                s['priority'] += priority
            except (KeyError, TypeError):
                s['priority'] = priority

            # Retrieve sound asset by name
            try:
                sound = self.machine.sounds[sound_name]
            except KeyError:
                self.machine.log.error("SoundPlayer: The specified sound "
                                       "does not exist ('{}').".format(sound_name))
                return

            s.update(kwargs)

            action = s['action'].lower()
            del s['action']

            track = self.machine.sound_system.audio_interface.get_track_by_name(s.get('track') or sound.track)
            if track is None:
                self.machine.log.error("SoundPlayer: The specified track ('{}') "
                                       "does not exist. Unable to perform '{}' action "
                                       "on sound '{}'."
                                       .format(s['track'], action, sound_name))
                return

            # Determine action to perform
            if action == 'play':
                track.play_sound(sound, context, s)

            elif action == 'stop':
                if 'fade_out' in s:
                    track.stop_sound(sound, s['fade_out'])
                else:
                    track.stop_sound(sound)

            elif action == 'stop_looping':
                track.stop_sound_looping(sound)

            elif action == 'load':
                sound.load()

            elif action == 'unload':
                sound.unload()

            else:
                self.machine.log.error("SoundPlayer: The specified action "
                                       "is not valid ('{}').".format(action))

    def get_express_config(self, value):
        """ express config for sounds is simply a string (sound name)"""
        return dict(sound=value)

    # pylint: disable=too-many-branches
    def validate_config(self, config):
        """Validates the sound_player: section of a config file (either a
        machine-wide config or a mode config).

        Args:
            config: A dict of the contents of the sound_player section
            from the config file. It's assumed that keys are event names, and
            values are settings for what the sound_player should do when that
            event is posted.

        Returns: A dict a validated entries.

        This method overrides the base method since the sound_player has
        unique options.

        """
        # first, we're looking to see if we have a string, a list, or a dict.
        # if it's a dict, we look to see whether we have the name of some sound

        validated_config = dict()

        for event, settings in config.items():
            validated_config[event] = dict()
            validated_config[event]['sounds'] = dict()

            if not isinstance(settings, dict):
                settings = {settings: dict()}

            if 'track' in settings:
                track = settings['track']

                if self.machine.sound_system.audio_interface.get_track_type(track) != "standard":
                    raise ValueError("SoundPlayer: An invalid audio track '{}' is specified for event '{}' "
                                     "(only standard audio tracks are supported).".format(track, event))

            for sound, sound_settings in settings.items():

                # Now check to see if all the settings are valid
                # sound settings. If not, assume it's a single sound settings.
                if isinstance(sound_settings, dict):
                    for key in sound_settings:
                        if key not in self.machine.config_validator.get_config_spec()['sound_player']:
                            break

                    validated_config[event]['sounds'].update(
                        self._validate_config_item(sound, sound_settings))

        return validated_config

    def clear_context(self, context):
        """Stop all sounds from this context."""
        self.machine.log.debug("SoundPlayer: Clearing context - applying mode_end_action for all active sounds")

        for index in range(self.machine.sound_system.audio_interface.get_track_count()):
            track = self.machine.sound_system.audio_interface.get_track(index)
            if track.type == "standard":
                track.clear_context(context)


mc_player_cls = McSoundPlayer



from mpfmc.core.audio.audio_exception import AudioException
from mpfmc.core.config_collection import ConfigCollection


class SoundLoopSetCollection(ConfigCollection):

    config_section = 'sound_loop_sets'
    collection = 'sound_loop_sets'
    class_label = 'SoundLoopSets'

    def __init__(self, mc, collection, config_section):
        super().__init__(mc, collection, config_section)

        self._validate_handler = None

    def create_entries(self, config: dict, **kwargs) -> None:
        # config is localized to this collection's section
        del kwargs

        # No need to create entries if sound system is not enabled
        if self.mc.sound_system is None or self.mc.sound_system.audio_interface is None:
            self.machine.log.info("Unable to create sound_loop_sets - sound system is not available")
            return

        for name, settings in config.items():
            try:
                self[name] = self.process_config(settings)
            except (AudioException, ValueError) as ex:
                raise ValueError("An error occurred while processing the '{}' entry in "
                                 "the sound_loop_sets config collection: {}".format(name, ex))

        # Validation of referenced sound assets must be completed after all
        # assets have been loaded (can use the init_done event for that)
        self._validate_handler = self.mc.events.add_handler("init_done", self._validate_sound_assets)

    def validate_entries_from_root_config(self, **kwargs):
        """Do nothing here."""

    def process_config(self, config: dict) -> dict:
        # processes the 'sound_loop_sets' section of a config file to populate the
        # mc.sound_loop_sets dict.

        # config is localized to 'sound_loop_sets' section
        return self.process_loop_set(config)

    def process_loop_set(self, config: dict) -> dict:
        # config is localized to a single sound loop set settings within a list

        self.mc.config_validator.validate_config('sound_loop_sets', config)

        # Clamp volume between 0 and 1
        if 'volume' in config and config['volume']:
            if config['volume'] < 0:
                config['volume'] = 0
            elif config['volume'] > 1:
                config['volume'] = 1

        # Validate optional layers
        if 'layers' in config:
            for layer in config["layers"]:
                self.mc.config_validator.validate_config('sound_loop_sets:layers', layer)

                # Clamp layer volume between 0 and 1
                if layer['volume'] < 0:
                    layer['volume'] = 0
                elif layer['volume'] > 1:
                    layer['volume'] = 1

        return config

    def _validate_sound_assets(self, **kwargs) -> None:
        """Validate the referenced sound assets in the loop set layers.

        Notes:
            This must be performed after all the sound assets have been loaded.
        """
        del kwargs
        if self._validate_handler:
            self.mc.events.remove_handler(self._validate_handler)

        if hasattr(self.mc, 'sounds'):
            for name, config in self.items():
                # Validate sound setting (make sure only valid sound assets are referenced)
                if config["sound"] not in self.mc.sounds:
                    raise ValueError("The '{}' sound_loop_set references an invalid sound asset "
                                     "name '{}' in its sound setting".format(name, config["sound"]))
                if self.mc.sounds[config["sound"]].streaming:
                    raise ValueError("The '{}' sound_loop_set references a streaming sound asset "
                                     "'{}' in its sound setting (only in-memory sounds are "
                                     "supported in loop sets)".format(name, config["sound"]))

                # Validate sound settings in layers (make sure only valid sound assets are referenced)
                for layer in config["layers"]:
                    if layer["sound"] not in self.mc.sounds:
                        raise ValueError("The '{}' sound_loop_set references an invalid sound asset "
                                         "name '{}' in one of its layers".format(name, layer["sound"]))
                    if self.mc.sounds[layer["sound"]].streaming:
                        raise ValueError("The '{}' sound_loop_set references a streaming sound asset "
                                         "'{}' in one of its layers (only in-memory sounds are "
                                         "supported in loop sets)".format(name, layer["sound"]))


collection_cls = SoundLoopSetCollection



# ---------------------------------------------------------------------------
#    Request Message types
# ---------------------------------------------------------------------------

cdef enum RequestMessage:
    request_sound_play = 1                # Request to play a sound
    request_sound_play_when_finished = 2  # Request to play a sound when the current one is finished
    request_sound_replace = 3             # Request to play a sound that replaces a sound in progress
    request_sound_stop = 4                # Request to stop a sound that is playing
    request_sound_stop_looping = 5        # Request to stop looping a sound that is playing


ctypedef struct RequestMessageDataPlaySound:
    SoundSample *sample
    Uint8 volume
    int loops
    int priority
    Uint32 start_at_position
    Uint32 fade_in_duration
    Uint32 fade_out_duration
    int marker_count
    GArray *markers
    bint sound_has_ducking
    DuckingSettings ducking_settings

ctypedef struct RequestMessageDataStopSound:
    Uint32 fade_out_duration
    Uint32 ducking_release_duration

ctypedef union RequestMessageData:
    RequestMessageDataPlaySound play
    RequestMessageDataStopSound stop

ctypedef struct RequestMessageContainer:
    RequestMessage message
    Uint64 sound_id
    Uint64 sound_instance_id
    int player
    RequestMessageData data





""" Contains the Mode parent class for the Media Controller"""

import logging
from collections import namedtuple

RemoteMethod = namedtuple('RemoteMethod',
                          'method config_section kwargs priority',
                          )
"""RemotedMethod is used by other modules that want to register a method to
be called on mode_start or mode_stop.
"""


# pylint: disable-msg=too-many-instance-attributes
class Mode:
    """Parent class for in-game mode code."""

    __slots__ = ["mc", "config", "name", "path", "asset_paths", "log", "priority", "_active", "stop_methods",
                 "start_callback", "stop_callback", "event_handlers", "target"]

    # pylint: disable-msg=too-many-arguments
    def __init__(self, mc, config, name, path, asset_paths):
        self.mc = mc
        self.config = config
        self.name = name.lower()
        self.path = path
        self.asset_paths = asset_paths

        self.log = logging.getLogger('Mode.' + name)

        self.priority = 0
        self._active = False
        self.stop_methods = list()
        self.start_callback = None
        self.stop_callback = None
        self.event_handlers = set()
        self.target = None

        if 'mode' in self.config:
            self.configure_mode_settings(config['mode'])

    def __repr__(self):
        return '<Mode.{}>'.format(self.name)

    @property
    def active(self):
        return self._active

    @active.setter
    def active(self, active):
        if self._active != active:
            self._active = active
            self.mc.mode_controller.active_change(self, self._active)

    def configure_mode_settings(self, config):
        """Processes this mode's configuration settings from a config
        dictionary.
        """

        if not ('priority' in config and isinstance(config['priority'], int)):
            config['priority'] = 0

        try:
            self.target = self.mc.targets[config['target']]
        except KeyError:
            self.target = self.mc.targets['default']

        self.config['mode'] = config

    def start(self, mode_priority=None, callback=None, **kwargs):
        """Starts this mode.

        Args:
            mode_ priority: Integer value of what you want this mode to run at.
                If you don't specify one, it will use the "Mode: priority"
                setting from this mode's configuration file.
            **kwargs: Catch-all since this mode might start from events with
                who-knows-what keyword arguments.

        Warning: You can safely call this method, but do not override it in
        your
        mode code. If you want to write your own mode code by subclassing Mode,
        put whatever code you want to run when this mode starts in the
        mode_start method which will be called automatically.

        """

        # todo might want to use this
        del callback
        del kwargs

        if isinstance(mode_priority, int):
            self.priority = mode_priority
        else:
            self.priority = self.config['mode']['priority']

        self.log.debug('Mode Start. Priority: %s', self.priority)

        self.active = True

        for item in self.mc.mode_controller.start_methods:
            if item.config_section in self.config or not item.config_section:
                self.stop_methods.append(item.method(
                    config=self.config.get(item.config_section),
                    priority=self.priority,
                    mode=self,
                    **item.kwargs))

    @staticmethod
    def is_game_mode() -> bool:
        """Return false since mc does not have game modes."""
        return False

    def stop(self, callback=None, **kwargs):
        """Stops this mode.

        Args:
            **kwargs: Catch-all since this mode might start from events with
                who-knows-what keyword arguments.

        Warning: You can safely call this method, but do not override it in
        your
        mode code. If you want to write your own mode code by subclassing Mode,
        put whatever code you want to run when this mode stops in the
        mode_stop method which will be called automatically.

        """

        # todo might want to use this?
        del callback

        del kwargs

        self.log.debug('Mode Stop.')

        self.priority = 0
        self.active = False

        for callback_func in self.mc.mode_controller.stop_methods:
            callback_func[0](self)

        for item in self.stop_methods:
            item[0](item[1])

        self.stop_methods = list()



from kivy.uix.effectwidget import EffectBase
from kivy.properties import NumericProperty


class GainEffect(EffectBase):
    """GLSL effect to apply apply a gain (brightness) adjustment to a texture.

    Args:
        gain: Float which adjusts the values. Default is 1.0 which has no
            effect, higher or lower values are multiplied by each color
            channel.

    """
    gain = NumericProperty(1.0)
    '''
    Sets the gain factor which is multiplied by each color channel.

    gain is a :class:`~kivy.properties.NumericProperty` and
    defaults to 1.0 (which has no effect).
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.do_glsl()

    def on_gain(self, *args):
        self.do_glsl()

    def do_glsl(self):
        self.glsl = gain_glsl.format(float(self.gain))


gain_glsl = '''
vec4 effect(vec4 color, sampler2D texture, vec2 tex_coords, vec2 coords)
{{
vec4 outColor = vec4(color.x * {0}, color.y * {0}, color.z * {0}, 1.0);
return outColor;
}}
'''

effect_cls = GainEffect
name = 'gain'



from mpfmc.core.mc_custom_code import McCustomCode


class MyCode(McCustomCode):

    def on_connect(self, **kwargs):
        self.add_mpf_event_handler("test_event", self._my_handler)

    def _my_handler(self, **kwargs):
        del kwargs
        self.post_event_to_mpf_and_mc("my_return_event")



from mpfmc.uix.display import Display, DisplayOutput
from mpfmc.tests.MpfMcTestCase import MpfMcTestCase


class TestDisplay(MpfMcTestCase):
    def get_machine_path(self):
        return 'tests/machine_files/display'

    def get_config_file(self):
        return 'test_display.yaml'

    def test_display(self):
        # Make sure nested multiple displays are loaded properly and are centered
        self.assertEqual(self.mc.root_window.size, (800, 600))
        self.assertIn('window', self.mc.displays)
        self.assertTrue(isinstance(self.mc.displays['window'], Display))
        self.assertEqual(self.mc.displays['window'].size, [600, 200])
        self.assertIsInstance(self.mc.displays['window'].parent, DisplayOutput)
        self.assertEqual(self.mc.displays['window'].parent.pos, (0, 167))

        self.assertIn('dmd', self.mc.displays)
        self.assertTrue(isinstance(self.mc.displays['dmd'], Display))
        self.assertEqual(self.mc.displays['dmd'].size, [128, 32])
        self.assertIsInstance(self.mc.displays['dmd'].parent, DisplayOutput)
        self.assertEqual(self.mc.displays['dmd'].parent.pos, (2, 0))

        self.assertEqual(self.mc.targets['window'], self.mc.displays['window'])
        self.assertEqual(self.mc.targets['dmd'], self.mc.displays['dmd'])
        self.assertEqual(self.mc.targets['default'], self.mc.displays['window'])

        self.assertEqual(self.mc.displays['window'].current_slide_name, 'window_slide_1')
        self.assertEqual(self.mc.displays['dmd'].current_slide_name, 'asset_status')



"""Widget showing a rectangle."""
from typing import Optional
from kivy.graphics import Rectangle as KivyRectangle, RoundedRectangle
from kivy.graphics.context_instructions import Color, Rotate, Scale
from kivy.properties import NumericProperty

from mpfmc.uix.widget import Widget

MYPY = False
if MYPY:   # pragma: no cover
    from mpfmc.core.mc import MpfMc


class Rectangle(Widget):

    """Widget showing a rectangle."""

    widget_type_name = 'Rectangle'
    animation_properties = ('x', 'y', 'width', 'height', 'color', 'opacity', 'corner_radius',
                            'rotation', 'scale')

    def __init__(self, mc: "MpfMc", config: dict, key: Optional[str] = None, **kwargs) -> None:
        del kwargs
        super().__init__(mc=mc, config=config, key=key)

        # Bind to all properties that when changed need to force
        # the widget to be redrawn
        self.bind(pos=self._draw_widget,
                  size=self._draw_widget,
                  color=self._draw_widget,
                  rotation=self._draw_widget,
                  scale=self._draw_widget,
                  corner_radius=self._draw_widget,
                  corner_segments=self._draw_widget)

        self._draw_widget()

    def _draw_widget(self, *args) -> None:
        """Establish the drawing instructions for the widget."""
        del args

        if self.canvas is None:
            return

        anchor = (self.x - self.anchor_offset_pos[0], self.y - self.anchor_offset_pos[1])
        self.canvas.clear()

        with self.canvas:
            Color(*self.color)
            Rotate(angle=self.rotation, origin=anchor)
            Scale(self.scale).origin = anchor

            if self.corner_radius > 0:
                RoundedRectangle(pos=self.pos, size=self.size,
                                 radius=(self.corner_radius,
                                         self.corner_radius),
                                 segments=self.corner_segments)
            else:
                KivyRectangle(pos=self.pos, size=self.size)

    #
    # Properties
    #

    corner_radius = NumericProperty(0)
    '''Specifies the radius of the round corners of the rectangle.
     Defaults to 0.
    '''

    corner_segments = NumericProperty(10)
    '''Defines how many segments will be used for drawing the round
    corners. The drawing will be smoother if you have many segments.
    '''

    rotation = NumericProperty(0)
    '''Rotation angle value of the widget.

    :attr:`rotation` is an :class:`~kivy.properties.NumericProperty` and defaults to
    0.
    '''

    scale = NumericProperty(1.0)
    '''Scale value of the widget.

    :attr:`scale` is an :class:`~kivy.properties.NumericProperty` and defaults to
    1.0.
    '''


widget_classes = [Rectangle]



import logging

from mpfmc.tests.MpfMcTestCase import MpfMcTestCase
from unittest.mock import MagicMock, ANY
from mpfmc.widgets.video import VideoWidget

try:
    from mpfmc.core.audio import SoundSystem
    from mpfmc.assets.sound import SoundStealingMethod
except ImportError:
    SoundSystem = None
    SoundStealingMethod = None
    logging.warning("mpfmc.core.audio library could not be loaded. Audio "
                    "features will not be available")


class TestAudioGStreamer(MpfMcTestCase):
    """
    Tests the GStreamer audio features in the media controller.  The core audio library is a
    custom extension library written in Cython that interfaces with the SDL2 and
    SDL_Mixer libraries.
    """
    def get_machine_path(self):
        return 'tests/machine_files/audio'

    def get_config_file(self):
        return 'test_audio_gstreamer.yaml'

    def test_loading_while_playing_video(self):
        """ Tests loading a sound file while playing a video (both using gstreamer) """

        if SoundSystem is None or self.mc.sound_system is None:
            log = logging.getLogger('TestAudio')
            log.warning("Sound system is not enabled - skipping audio tests")
            self.skipTest("Sound system is not enabled")

        self.assertIsNotNone(self.mc.sound_system)
        interface = self.mc.sound_system.audio_interface
        if interface is None:
            log = logging.getLogger('TestAudio')
            log.warning("Sound system audio interface could not be loaded - skipping audio tests")
            self.skipTest("Sound system audio interface could not be loaded")

        self.assertIsNotNone(interface)

        # Mock BCP send method
        self.mc.bcp_processor.send = MagicMock()
        self.mc.bcp_processor.enabled = True

        # Make sure the low quality test video exists
        self.assertIn('mpf_video_small_test', self.mc.videos)

        self.mc.events.post('show_slide1')
        self.advance_time()

        video_widget = self.mc.targets['default'].current_slide.widgets[0].widget

        self.assertEqual(video_widget.state, 'play')
        self.assertTrue(video_widget.video.loaded)

        self.mc.events.post('play_sound_sfx_028')
        self.advance_real_time(1)
        self.mc.events.post('play_city_loop')

        self.advance_real_time(1)

        self.mc.events.post('play_sound_text')
        self.advance_real_time(0.35)
        self.mc.events.post('play_sound_text')
        self.advance_real_time(6)

        self.mc.bcp_processor.send.assert_any_call('trigger', sound_instance=ANY, name='text_sound_played')
        self.mc.bcp_processor.send.assert_any_call('trigger', name='test_video_played')
        self.mc.bcp_processor.send.assert_any_call('trigger', name='test_video_stopped')

    def test_retrigger_streamed_sound(self):
        """ Tests retriggering a streamed sound (play, stop, play) (using gstreamer) """

        if SoundSystem is None or self.mc.sound_system is None:
            log = logging.getLogger('TestAudio')
            log.warning("Sound system is not enabled - skipping audio tests")
            self.skipTest("Sound system is not enabled")

        self.assertIsNotNone(self.mc.sound_system)
        interface = self.mc.sound_system.audio_interface
        if interface is None:
            log = logging.getLogger('TestAudio')
            log.warning("Sound system audio interface could not be loaded - skipping audio tests")
            self.skipTest("Sound system audio interface could not be loaded")

        self.assertIsNotNone(interface)
        track_music = interface.get_track_by_name("music")
        self.assertIsNotNone(track_music)

        # Mock BCP send method
        self.mc.bcp_processor.send = MagicMock()
        self.mc.bcp_processor.enabled = True

        self.mc.events.post('play_city_loop')

        self.advance_real_time(1)
        self.assertTrue(track_music.sound_is_playing(self.mc.sounds['city_loop']))

        self.mc.events.post('stop_city_loop')
        self.advance_real_time(0.25)
        self.assertFalse(track_music.sound_is_playing(self.mc.sounds['city_loop']))
        self.advance_real_time(0.25)
        self.mc.events.post('play_city_loop')
        self.advance_real_time(1)
        self.assertTrue(track_music.sound_is_playing(self.mc.sounds['city_loop']))
        self.advance_real_time(1)


